import pytest

from datarobot import Client
from datarobot.client import get_client


def test_missing_token_in_env_fails(clean_env):
    clean_env.setenv('DATAROBOT_ENDPOINT', 'https://home.com')
    with pytest.raises(ValueError):
        get_client()


def test_missing_endpoint_in_config_fails():
    with pytest.raises(ValueError):
        Client(token='a-token')


def test_missing_token_in_config_fails():
    with pytest.raises(ValueError):
        Client(endpoint='https://host_name.com')


@pytest.mark.usefixtures('no_version_check')
def test_file_config_with_connect_timeout(clean_user_config, default_config_token,
                                          default_config_endpoint):
    fcontent = 'endpoint: {}\ntoken: {}\nconnect_timeout: 23'.format(
        default_config_endpoint,
        default_config_token
    )
    clean_user_config.write(fcontent)
    client = get_client()
    assert client.connect_timeout == 23


def test_code_config_with_connect_timeout(code_config_token, code_config_endpoint):
    Client(token=code_config_token, endpoint=code_config_endpoint,
           connect_timeout=24)
    c = get_client()
    assert c.connect_timeout == 24



import trafaret as t

from datarobot.client import get_client, staticproperty
from datarobot.models.api_object import APIObject

from ..utils import encode_utf8_if_py2, from_api, parse_time


_data_store_params_converter = t.Dict({
    t.Key('driver_id'): t.String(),
    t.Key('jdbc_url'): t.String()
}).ignore_extra('*')


class DataStoreParameters(object):
    def __init__(self, driver_id, jdbc_url):
        _data_store_params_converter.check({
            'driver_id': driver_id,
            'jdbc_url': jdbc_url
        })
        self.driver_id = driver_id
        self.jdbc_url = jdbc_url

    def collect_payload(self):
        return {
            'driver_id': self.driver_id,
            'jdbc_url': self.jdbc_url
        }


class DataStore(APIObject):
    """ A data store. Represents database

    Attributes
    ----------
    id : str
        the id of the data store.
    data_store_type : str
        the type of data store.
    canonical_name : str
        the user-friendly name of the data store.
    creator : str
        the id of the user who created the data store.
    updated : datetime.datetime
        the time of the last update
    params : DataStoreParameters
        a list specifying data store parameters.
    """
    _path = 'externalDataStores/'
    _client = staticproperty(get_client)
    _converter = t.Dict({
        t.Key('id', optional=True) >> 'data_store_id': t.String(),
        t.Key('type') >> 'data_store_type': t.String(),
        t.Key('canonical_name'): t.String(),
        t.Key('creator'): t.String(),
        t.Key('params'): _data_store_params_converter,
        t.Key('updated'): parse_time,
    }).ignore_extra('*')

    def __init__(self, data_store_id=None, data_store_type=None, canonical_name=None, creator=None,
                 updated=None, params=None):
        self._id = data_store_id
        self._type = data_store_type
        self.canonical_name = canonical_name
        self._creator = creator
        self._updated = updated
        self.params = params

    @classmethod
    def list(cls):
        """
        Returns list of available data stores.

        Returns
        -------
        data_stores : list of DataStore instances
            contains a list of available data stores.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_stores = dr.DataStore.list()
            >>> data_stores
            [DataStore('Demo'), DataStore('Airlines')]
        """
        r_data = cls._client.get(cls._path).json()
        return [cls.from_server_data(item) for item in r_data['data']]

    @classmethod
    def get(cls, data_store_id):
        """
        Gets the data store.

        Parameters
        ----------
        data_store_id : str
            the identifier of the data store.

        Returns
        -------
        data_store : DataStore
            the required data store.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5a8ac90b07a57a0001be501e')
            >>> data_store
            DataStore('Demo')
        """
        return cls.from_location('{}{}/'.format(cls._path, data_store_id))

    @classmethod
    def create(cls, data_store_type, canonical_name, driver_id, jdbc_url):
        """
        Creates the data store.

        Parameters
        ----------
        data_store_type : str
            the type of data store.
        canonical_name : str
            the user-friendly name of the data store.
        driver_id : str
            the identifier of the DataDriver.
        jdbc_url : str
             the full JDBC url, for example `jdbc:postgresql://my.dbaddress.org:5432/my_db`.

        Returns
        -------
        data_store : DataStore
            the created data store.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.create(
            ...     data_store_type='jdbc',
            ...     canonical_name='Demo DB',
            ...     driver_id='5a6af02eb15372000117c040',
            ...     jdbc_url='jdbc:postgresql://my.db.address.org:5432/perftest'
            ... )
            >>> data_store
            DataStore('Demo DB')
        """
        payload = {
            'type': data_store_type,
            'canonicalName': canonical_name,
            'params': DataStoreParameters(
                driver_id=driver_id,
                jdbc_url=jdbc_url
            ).collect_payload()
        }
        return cls.from_server_data(
            cls._client.post(cls._path, data=payload).json()
        )

    def update(self, canonical_name=None, driver_id=None, jdbc_url=None):
        """
        Updates the data store.

        Parameters
        ----------
        canonical_name : str
            optional, the user-friendly name of the data store.
        driver_id : str
            optional, the identifier of the DataDriver.
        jdbc_url : str
            optional, the full JDBC url,
            for example `jdbc:postgresql://my.dbaddress.org:5432/my_db`.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5ad5d2afef5cd700014d3cae')
            >>> data_store
            DataStore('Demo DB')
            >>> data_store.update(canonical_name='Demo DB updated')
            >>> data_store
            DataStore('Demo DB updated')
        """
        payload = {
            'canonicalName': canonical_name or self.canonical_name,
            'params': DataStoreParameters(
                driver_id=driver_id or self.params.driver_id,
                jdbc_url=jdbc_url or self.params.jdbc_url
            ).collect_payload()
        }
        r_data = self._client.patch('{}{}/'.format(self._path, self.id), data=payload).json()
        self.canonical_name = r_data['canonicalName']
        self.params = DataStoreParameters(r_data['params']['driverId'], r_data['params']['jdbcUrl'])

    def delete(self):
        """ Removes the DataStore """
        self._client.delete('{}{}/'.format(self._path, self.id))

    def test(self, username, password):
        """
        Tests database connection.

        Parameters
        ----------
        username : str
            the username for database authentication.
        password : str
            the password for database authentication. The password is encrypted
            at server side and never saved / stored

        Returns
        -------
        message : dict
            message with status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5ad5d2afef5cd700014d3cae')
            >>> data_store.test(username='db_username', password='db_password')
            {'message': 'Connection successful'}
        """
        payload = {
            'user': username,
            'password': password
        }
        return self._client.post('{}{}/test/'.format(self._path, self.id), data=payload).json()

    def schemas(self, username, password):
        """
        Returns list of available schemas.

        Parameters
        ----------
        username : str
            the username for database authentication.
        password : str
            the password for database authentication. The password is encrypted
            at server side and never saved / stored

        Returns
        -------
        response : dict
            dict with database name and list of str - available schemas

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5ad5d2afef5cd700014d3cae')
            >>> data_store.schemas(username='db_username', password='db_password')
            {'catalog': 'perftest', 'schemas': ['demo', 'information_schema', 'public']}
        """
        payload = {
            'user': username,
            'password': password
        }
        return self._client.post('{}{}/schemas/'.format(self._path, self.id), data=payload).json()

    def tables(self, username, password, schema=None):
        """
        Returns list of available tables in schema.

        Parameters
        ----------
        username : str
            optional, the username for database authentication.
        password : str
            optional, the password for database authentication. The password is encrypted
            at server side and never saved / stored
        schema : str
            optional, the schema name.

        Returns
        -------
        response : dict
            dict with catalog name and tables info

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5ad5d2afef5cd700014d3cae')
            >>> data_store.tables(username='db_username', password='db_password', schema='demo')
            {'tables': [{'type': 'TABLE', 'name': 'diagnosis', 'schema': 'demo'}, {'type': 'TABLE',
            'name': 'kickcars', 'schema': 'demo'}, {'type': 'TABLE', 'name': 'patient',
            'schema': 'demo'}, {'type': 'TABLE', 'name': 'transcript', 'schema': 'demo'}],
            'catalog': 'perftest'}
        """
        payload = {
            'schema': schema,
            'user': username,
            'password': password
        }
        return self._client.post('{}{}/tables/'.format(self._path, self.id), data=payload).json()

    @classmethod
    def from_server_data(cls, data, keep_attrs=None):
        converted_data = cls._converter.check(from_api(data))
        params = converted_data.pop('params')
        converted_data['params'] = DataStoreParameters(params['driver_id'], params['jdbc_url'])
        return cls(**converted_data)

    def __repr__(self):
        return encode_utf8_if_py2(u'{}(\'{}\')'.format(self.__class__.__name__,
                                                       self.canonical_name or self.id))

    @property
    def id(self):
        return self._id

    @property
    def creator(self):
        return self._creator

    @property
    def type(self):
        return self._type

    @property
    def updated(self):
        return self._updated



import pytest
import responses

from datarobot import errors
from datarobot.utils.waiters import wait_for_async_resolution


@responses.activate
def test_success(client, async_url, async_success_callback, resolved_url):
    responses.add_callback(
        responses.GET,
        async_url,
        callback=async_success_callback,
        content_type='application/json'
    )
    result = wait_for_async_resolution(client, async_url)
    assert result == resolved_url


@responses.activate
def test_job_failure(client,
                     async_url,
                     async_job_failure_callback,
                     ):
    responses.add_callback(
        responses.GET,
        async_url,
        callback=async_job_failure_callback,
        content_type='application/json'
    )
    with pytest.raises(errors.AsyncProcessUnsuccessfulError):
        wait_for_async_resolution(client, async_url)


@responses.activate
def test_server_wtf(client,
                    async_url,
                    async_server_nonsense_callback,
                    ):
    responses.add_callback(
        responses.GET,
        async_url,
        callback=async_server_nonsense_callback,
        content_type='application/json'
    )
    with pytest.raises(errors.AsyncFailureError):
        wait_for_async_resolution(client, async_url)


@responses.activate
def test_server_timeout(client,
                        async_url,
                        async_not_finished_callback,
                        mock_async_time):
    responses.add_callback(
        responses.GET,
        async_url,
        callback=async_not_finished_callback,
        content_type='application/json'
    )
    mock_async_time.time.side_effect = (0, 0, 10**6)

    with pytest.raises(errors.AsyncTimeoutError):
        wait_for_async_resolution(client, async_url)



from pprint import pformat
from textwrap import dedent


class NoParametersFoundException(Exception):
    """No parameters were found that matched the specified filter"""
    pass


class NonUniqueParametersException(Exception):
    """Multiple parameters were found that matched the specified filter"""
    def __init__(self, keys, matching_params):
        """Construct a NonUniqueParametersException

        Params
        ------
        keys : dict
            Keys used for filtering
        matching_params : list(dict)
            Parameters that matched the specified filter
        """
        self.matching_params = matching_params

        # Strip non-identifying fields from params for display purposes,
        # to make the error message more concise
        non_identifying_fields = {'defaultValue', 'currentValue', 'constraints'}
        filtered_params = [
            {
                key: value
                for key, value in param.items()
                if key not in non_identifying_fields
            }
            for param in matching_params
        ]

        message = dedent("""\
            Multiple matching parameters found for the specified keys.
            Keys: {}
            Parameters:
            {}""").format(pformat(keys), pformat(filtered_params))
        super(NonUniqueParametersException, self).__init__(message)


class AdvancedTuningSession(object):
    """A session enabling users to configure and run advanced tuning for a model.

    Every model contains a set of one or more tasks.  Every task contains a set of
    zero or more parameters.  This class allows tuning the values of each parameter
    on each task of a model, before running that model.

    This session is client-side only and is not persistent.
    Only the final model, constructed when `run` is called, is persisted on the DataRobot server.

    Attributes
    ----------
    description : basestring
        Description for the new advance-tuned model.
        Defaults to the same description as the base model.
    """

    def __init__(self, model):
        """Initiate an Advanced Tuning session.

        Params
        ------
        model : datarobot.models.model.Model
        """
        self._new_values = {}

        self._model = model

        param_info = model.get_advanced_tuning_parameters()
        self._available_params = param_info['tuning_parameters']
        self.description = param_info.get('tuning_description')

    def _get_parameter_id(self, task_name=None, parameter_name=None, parameter_id=None):
        """Return the ID of the one parameter that matches the specified fields.

        Returns
        -------
        dict
            Parameter dict -- an entry in self._available_params
            (see docstring for `Model.get_advanced_tuning_params`)

        Raises
        ------
        NoParametersFoundException
            if no matching parameters are found.
        NonUniqueParametersException
            if multiple parameters matched the specified filtering criteria
        """
        filtered_params = (x for x in self._available_params)
        if parameter_id:
            # Should be unique but filter normally just in case it's ever not unique
            filtered_params = (x for x in filtered_params if x['parameter_id'] == parameter_id)
        if parameter_name:
            filtered_params = (x for x in filtered_params if x['parameter_name'] == parameter_name)
        if task_name:
            filtered_params = (x for x in filtered_params if x['task_name'] == task_name)

        filtered_params = list(filtered_params)

        if len(filtered_params) == 0:
            raise NoParametersFoundException(
                'No parameters found with task_name of {} and parameter_name of {}'.format(
                    repr(task_name) if task_name else '(unspecified)',
                    repr(parameter_name) if parameter_name else '(unspecified)'
                ))

        if len(filtered_params) > 1:
            key = {}
            if task_name:
                key['task_name'] = task_name
            if parameter_name:
                key['parameter_name'] = parameter_name
            if parameter_id:
                key['parameter_id'] = parameter_id
            raise NonUniqueParametersException(key, filtered_params)

        return filtered_params[0]['parameter_id']

    def get_task_names(self):
        """Get the list of task names that are available for this model

        Returns
        -------
        list(basestring)
            List of task names
        """
        return sorted(set(x['task_name'] for x in self._available_params))

    def get_parameter_names(self, task_name):
        """Get the list of parameter names available for a specific task

        Returns
        -------
        list(basestring)
            List of parameter names
        """
        return [
            x['parameter_name']
            for x in self._available_params
            if x['task_name'] == task_name
        ]

    def set_parameter(
            self, value,
            task_name=None,
            parameter_name=None,
            parameter_id=None):
        """Set the value of a parameter to be used

        The caller must supply enough of the optional arguments to this function
        to uniquely identify the parameter that is being set.
        For example, a less-common parameter name such as
        'building_block__complementary_error_function' might only be used once (if at all)
        by a single task in a model.  In which case it may be sufficient to simply specify
        'parameter_name'.  But a more-common name such as 'random_seed' might be used by
        several of the model's tasks, and it may be necessary to also specify 'task_name'
        to clarify which task's random seed is to be set.

        Parameters
        ----------
        task_name : basestring
            Name of the task whose parameter needs to be set
        parameter_name : basestring
            Name of the parameter to set
        parameter_id : basestring
            ID of the parameter to set
        value : int, float, list, or basestring
            New value for the parameter, with legal values determined by the parameter being set

        Raises
        ------
        NoParametersFoundException
            if no matching parameters are found.
        NonUniqueParametersException
            if multiple parameters matched the specified filtering criteria
        """
        parameter_id = self._get_parameter_id(
            task_name=task_name,
            parameter_name=parameter_name,
            parameter_id=parameter_id
        )

        self._new_values[parameter_id] = value

    def _add_value_to_param(self, param):
        """Given a 'param' dict, add a new user-specified value (if any) and return"""
        new_param = dict(param)
        new_param['value'] = self._new_values.get(param['parameter_id'])
        return new_param

    def get_parameters(self):
        """Returns the set of parameters available to this model

        The returned parameters have one additional key, "value", reflecting any new values that
        have been set in this AdvancedTuningSession.  When the session is run, "value" will be used,
        or if it is unset, "current_value".


        Returns
        -------
        parameters : dict
            "Parameters" dictionary, same as specified on `Model.get_advanced_tuning_params`.

        An additional field is added per parameter to the 'tuningParameters' list in the dictionary:

        value : int, float, list, or basestring
            The current value of the parameter.  `None` if none has been specified.
        """

        return {
            'tuning_description': self.description,
            'tuning_parameters': [
                self._add_value_to_param(param)
                for param in self._available_params
            ]
        }

    def run(self):
        """Submit this model for Advanced Tuning.

        Returns
        -------
        datarobot.models.modeljob.ModelJob
            The created job to build the model
        """
        return self._model.advanced_tune(self._new_values, self.description)



import json

import pytest
import responses
from datarobot import Model

from datarobot.models.confusion_chart import ConfusionChart


@pytest.fixture
def confusion_chart_validation_data():
    return {
        "data": {
            "classMetrics": [
                {
                    "actualCount": 9,
                    "className": "1",
                    "confusionMatrixOneVsAll": [
                            [15, 0], [0, 9]],
                    "f1": 1.0,
                    "precision": 1.0,
                    "predictedCount": 9,
                    "recall": 1.0,
                    "wasActualPercentages": [
                        {"otherClassName": "1", "percentage": 1.0},
                        {"otherClassName": "2", "percentage": 0.0},
                        {"otherClassName": "3", "percentage": 0.0}
                    ],
                    "wasPredictedPercentages": [
                        {"otherClassName": "1", "percentage": 1.0},
                        {"otherClassName": "2", "percentage": 0.0},
                        {"otherClassName": "3", "percentage": 0.0}
                    ]
                },
                {
                    "actualCount": 5,
                    "className": "2",
                    "confusionMatrixOneVsAll": [
                        [18, 1], [1, 4]
                    ],
                    "f1": 0.8000000000000002,
                    "precision": 0.8,
                    "predictedCount": 5,
                    "recall": 0.8,
                    "wasActualPercentages": [
                        {"otherClassName": "1", "percentage": 0.0},
                        {"otherClassName": "2", "percentage": 0.8},
                        {"otherClassName": "3", "percentage": 0.2}
                    ],
                    "wasPredictedPercentages": [
                        {"otherClassName": "1", "percentage": 0.0},
                        {"otherClassName": "2", "percentage": 0.8},
                        {"otherClassName": "3", "percentage": 0.2}
                    ]
                },
                {
                    "actualCount": 10,
                    "className": "3",
                    "confusionMatrixOneVsAll": [
                        [13, 1], [1, 9]
                    ],
                    "f1": 0.9,
                    "precision": 0.9,
                    "predictedCount": 10,
                    "recall": 0.9,
                    "wasActualPercentages": [
                        {"otherClassName": "1", "percentage": 0.0},
                        {"otherClassName": "2", "percentage": 0.1},
                        {"otherClassName": "3", "percentage": 0.9}
                    ],
                    "wasPredictedPercentages": [
                        {"otherClassName": "1", "percentage": 0.0},
                        {"otherClassName": "2", "percentage": 0.1},
                        {"otherClassName": "3", "percentage": 0.9}
                    ]
                }
            ],
            "confusionMatrix": [
                [9, 0, 0], [0, 4, 1], [0, 1, 9]
            ]
        },
        "source": "validation"
    }


@pytest.fixture
def confusion_chart_metadata_server_data():
    return {
        'source': 'validation',
        'classNames': ['1', '2', '3'],
        'totalMatrixSum': 24,
        'relevantClassesPositions': [[1]]
    }


def test_instantiation(confusion_chart_validation_data):
    confusion_chart_validation_data['data']['classes'] = ['1', '2', '3']
    confusion_chart = ConfusionChart.from_server_data(confusion_chart_validation_data)

    assert confusion_chart.classes == ['1', '2', '3']
    assert confusion_chart.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert confusion_chart.source == 'validation'


def test_future_proof(confusion_chart_validation_data):
    data_with_future_keys = dict(confusion_chart_validation_data, new_key='some future data')
    data_with_future_keys['data']['classMetrics'][0]['new_key'] = 'some future bin data'
    data_with_future_keys['data']['classes'] = ['1', '2', '3']
    confusion_chart = ConfusionChart.from_server_data(data_with_future_keys)

    assert confusion_chart.classes == ['1', '2', '3']
    assert confusion_chart.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]
    assert confusion_chart.source == 'validation'


@pytest.fixture
def confusion_chart_validation_data_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/confusionCharts/{}/'.format(
        project_id,
        model_id,
        'validation'
    )


@pytest.fixture
def confusion_chart_list_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/confusionCharts/'.format(
        project_id,
        model_id
    )


@pytest.fixture
def confusion_chart_metadata_response(project_url, model_id, confusion_chart_metadata_server_data):
    responses.add(
        responses.GET,
        '{}models/{}/confusionCharts/validation/metadata/'.format(project_url, model_id),
        json=confusion_chart_metadata_server_data, content_type='application/json'
    )


@responses.activate
@pytest.mark.usefixtures('client', 'confusion_chart_metadata_response')
def test_get_validation_confusion_chart(confusion_chart_validation_data,
                                        confusion_chart_validation_data_url,
                                        project_id, model_id):
    responses.add(
        responses.GET,
        confusion_chart_validation_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(confusion_chart_validation_data)
    )
    model = Model(id=model_id, project_id=project_id)
    cm = model.get_confusion_chart('validation')

    assert cm.source == confusion_chart_validation_data['source']
    assert cm.classes == ['1', '2', '3']
    assert cm.confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]


@responses.activate
@pytest.mark.usefixtures('client', 'confusion_chart_metadata_response')
def test_get_all_confusion_chart(confusion_chart_validation_data,
                                 confusion_chart_list_url,
                                 project_id, model_id):
    responses.add(
        responses.GET,
        confusion_chart_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [confusion_chart_validation_data]})
    )
    model = Model(id=model_id, project_id=project_id)
    cm_list = model.get_all_confusion_charts()

    assert len(cm_list) == 1
    assert cm_list[0].source == confusion_chart_validation_data['source']
    assert cm_list[0].classes == ['1', '2', '3']
    assert cm_list[0].confusion_matrix == [[9, 0, 0], [0, 4, 1], [0, 1, 9]]



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.model import Model
from ..utils import encode_utf8_if_py2


class ModelRecommendation(APIObject):
    """ A collection of information about a recommended model for a project.

    Attributes
    ----------
    project_id : str
        the id of the project the model belongs to
    model_id : str
        the id of the recommended model
    recommendation_type : str
        the type of model recommendation
    """

    _base_recommended_path_template = 'projects/{}/recommendedModels/'
    _converter = t.Dict({
        t.Key('project_id'): t.String,
        t.Key('model_id'): t.String,
        t.Key('recommendation_type'): t.String
    }).ignore_extra('*')

    def __init__(self, project_id, model_id, recommendation_type):
        self.project_id = project_id
        self.model_id = model_id
        self.recommendation_type = recommendation_type

    def __repr__(self):
        return encode_utf8_if_py2(u'ModelRecommendation({}, {}, {})'.format(
            self.project_id,
            self.model_id,
            self.recommendation_type
        ))

    @classmethod
    def get(cls, project_id, recommendation_type=None):
        """
        Retrieves the default or specified by recommendation_type recommendation.

        Parameters
        ----------
        project_id : str
            The project's id.
        recommendation_type : enums.RECOMMENDED_MODEL_TYPE
            The type of recommendation to get. If None, returns the default recommendation.

        Returns
        -------
        recommended_model : ModelRecommendation

        """
        if recommendation_type is None:
            url = cls._base_recommended_path_template.format(project_id) + 'recommendedModel/'
            return cls.from_location(url)
        else:
            recommendations = cls.get_all(project_id)
            return cls.get_recommendation(recommendations, recommendation_type)

    @classmethod
    def get_all(cls, project_id):
        """
        Retrieves all of the current recommended models for the project.


        Parameters
        ----------
        project_id : str
            The project's id.

        Returns
        -------
        recommended_models : list of ModelRecommendation
        """
        url = cls._base_recommended_path_template.format(project_id)
        response = ModelRecommendation._server_data(url)
        return [ModelRecommendation.from_server_data(data) for data in response]

    @classmethod
    def get_recommendation(cls, recommended_models, recommendation_type):
        """
        Returns the model in the given list with the requested type.


        Parameters
        ----------
        recommended_models : list of ModelRecommendation
        recommendation_type : enums.RECOMMENDED_MODEL_TYPE
            the type of model to extract from the recommended_models list

        Returns
        -------
        recommended_model : ModelRecommendation or None if no model with the requested type exists
        """

        return next(
            (
                model for model in recommended_models
                if model.recommendation_type == recommendation_type
            ),
            None
        )

    def get_model(self):
        """
        Returns the Model associated with this ModelRecommendation.

        Returns
        -------
        recommended_model : Model
        """
        return Model.get(self.project_id, self.model_id)



import trafaret as t
import numpy as np

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2


class RocCurve(APIObject):
    """ ROC curve data for model.

    Attributes
    ----------
    source : str
        ROC curve data source. Can be 'validation', 'crossValidation' or 'holdout'.
    roc_points : list of dict
        List of precalculated metrics associated with thresholds for ROC curve.
    negative_class_predictions : list of float
        List of predictions from example for negative class
    positive_class_predictions : list of float
        List of predictions from example for positive class
    """
    _converter = t.Dict({
        t.Key('source'): t.String,
        t.Key('roc_points'): t.List(t.Dict({
            t.Key('accuracy'): t.Float,
            t.Key('f1_score'): t.Float,
            t.Key('false_negative_score'): t.Int,
            t.Key('true_negative_score'): t.Int,
            t.Key('true_positive_score'): t.Int,
            t.Key('false_positive_score'): t.Int,
            t.Key('true_negative_rate'): t.Float,
            t.Key('false_positive_rate'): t.Float,
            t.Key('true_positive_rate'): t.Float,
            t.Key('matthews_correlation_coefficient'): t.Float,
            t.Key('positive_predictive_value'): t.Float,
            t.Key('negative_predictive_value'): t.Float,
            t.Key('threshold'): t.Float,
        }).ignore_extra('*')),
        t.Key('negative_class_predictions'): t.List(t.Float),
        t.Key('positive_class_predictions'): t.List(t.Float),
    }).ignore_extra('*')

    def __init__(self, source, roc_points,
                 negative_class_predictions, positive_class_predictions):
        self.source = source
        self.roc_points = roc_points
        self.negative_class_predictions = negative_class_predictions
        self.positive_class_predictions = positive_class_predictions

    def __repr__(self):
        return encode_utf8_if_py2(u'RocCurve({})'.format(self.source))

    @staticmethod
    def _validate_threshold(threshold):
        if threshold > 1 or threshold < 0:
            raise ValueError('threshold must be from [0, 1] interval')

    def estimate_threshold(self, threshold):
        """ Return metrics estimation for given threshold.

        Parameters
        ----------
        threshold : float from [0, 1] interval
            Threshold we want estimation for

        Returns
        -------
        dict
            Dictionary of estimated metrics in form of {metric_name: metric_value}.
            Metrics are 'accuracy', 'f1_score', 'false_negative_score', 'true_negative_score',
            'true_negative_rate', 'matthews_correlation_coefficient', 'true_positive_score',
            'positive_predictive_value', 'false_positive_score', 'false_positive_rate',
            'negative_predictive_value', 'true_positive_rate'.

        Raises
        ------
        ValueError
            Given threshold isn't from [0, 1] interval
        """
        self._validate_threshold(threshold)
        for roc_point in self.roc_points:
            if np.isclose(roc_point['threshold'], threshold):
                estimated_roc_point = roc_point
                break
        else:
            # if no exact match - pick closest ROC point with bigger threshold
            roc_points_with_bigger_threshold = [roc_point for roc_point in self.roc_points
                                                if roc_point['threshold'] > threshold]
            estimated_roc_point = sorted(roc_points_with_bigger_threshold,
                                         key=lambda rp: rp['threshold'])[0]
        return estimated_roc_point

    def get_best_f1_threshold(self):
        """ Return value of threshold that corresponds to max F1 score.
        This is threshold that will be preselected in DataRobot when you open "ROC curve" tab.

        Returns
        -------
        float
            Threhold with best F1 score.
        """
        roc_point_with_best_f1 = max(self.roc_points,
                                     key=lambda roc_point: roc_point['f1_score'])
        return roc_point_with_best_f1['threshold']



import six
import trafaret as t

from datarobot.client import get_client, staticproperty
from datarobot.utils import from_api


class APIObject(object):
    _client = staticproperty(get_client)
    _converter = t.Dict({}).allow_extra('*')

    @classmethod
    def _fields(cls):
        return {k.to_name or k.name for k in cls._converter.keys}

    @classmethod
    def from_data(cls, data):
        checked = cls._converter.check(data)
        safe_data = cls._filter_data(checked)
        return cls(**safe_data)

    @classmethod
    def from_location(cls, path, keep_attrs=None):
        server_data = cls._server_data(path)
        return cls.from_server_data(server_data, keep_attrs=keep_attrs)

    @classmethod
    def from_server_data(cls, data, keep_attrs=None):
        """
        Instantiate an object of this class using the data directly from the server,
        meaning that the keys may have the wrong camel casing

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        keep_attrs : list
            List of the dotted namespace notations for attributes to keep within the
            object structure even if their values are None
        """
        case_converted = from_api(data, keep_attrs=keep_attrs)
        return cls.from_data(case_converted)

    @classmethod
    def _filter_data(cls, data):
        fields = cls._fields()
        return {key: value
                for key, value in six.iteritems(data)
                if key in fields}

    @classmethod
    def _safe_data(cls, data):
        return cls._filter_data(cls._converter.check(from_api(data, do_recursive=False)))

    @classmethod
    def _server_data(cls, path):
        return cls._client.get(path).json()



import trafaret as t

from .. import errors
from .. import enums
from .job import AbstractSpecificJob
from ..utils import from_api, retry


def wait_for_async_model_creation(project_id, model_job_id, max_wait=enums.DEFAULT_MAX_WAIT):
    """
    Given a Project id and ModelJob id poll for status of process
    responsible for model creation until model is created.

    Parameters
    ----------
    project_id : str
        The identifier of the project

    model_job_id : str
        The identifier of the ModelJob

    max_wait : int, optional
        Time in seconds after which model creation is considered
        unsuccessful

    Returns
    -------
    model : Model
        Newly created model

    Raises
    ------
    AsyncModelCreationError
        Raised if status of fetched ModelJob object is ``error``
    AsyncTimeoutError
        Model wasn't created in time, specified by ``max_wait`` parameter
    """
    # Note: Don't need this in 3.0. (Use `Job.get_result_when_complete` method instead.)
    for _ in retry.wait(max_wait):
        try:
            model_job = ModelJob.get(project_id, model_job_id)
        except errors.PendingJobFinished:
            return ModelJob.get_model(project_id, model_job_id)
        if model_job.status == enums.MODEL_JOB_STATUS.ERROR:
            e_msg = 'Model creation unsuccessful'
            raise errors.AsyncModelCreationError(e_msg)

    timeout_msg = 'Model creation timed out in {} seconds'.format(max_wait)
    raise errors.AsyncTimeoutError(timeout_msg)


class ModelJob(AbstractSpecificJob):

    """ Tracks asynchronous work being done within a project

    Attributes
    ----------
    id : int
        the id of the job
    project_id : str
        the id of the project the job belongs to
    status : str
        the status of the job - will be one of ``datarobot.enums.QUEUE_STATUS``
    job_type : str
        what kind of work the job is doing - will be 'model' for modeling jobs
    is_blocked : bool
        if true, the job is blocked (cannot be executed) until its dependencies are resolved
    sample_pct : float
        the percentage of the project's dataset used in this modeling job
    model_type : str
        the model this job builds (e.g. 'Nystroem Kernel SVM Regressor')
    processes : list of str
        the processes used by the model
    featurelist_id : str
        the id of the featurelist used in this modeling job
    blueprint : Blueprint
        the blueprint used in this modeling job
    """

    _extra_fields = frozenset(['sample_pct',
                               'model_type',
                               'processes',
                               'featurelist_id'])

    _converter_extra = t.Dict({
        t.Key('sample_pct', optional=True): t.Float,
        t.Key('model_type', optional=True): t.String,
        t.Key('processes', optional=True): t.List(t.String),
        t.Key('featurelist_id', optional=True): t.String,
    })

    def __repr__(self):
        return 'ModelJob({}, status={})'.format(self.model_type,
                                                self.status)  # pragma:no cover

    def __init__(self, data, completed_resource_url=None):
        super(ModelJob, self).__init__(data, completed_resource_url=completed_resource_url)
        from . import Blueprint

        data = from_api(data)

        # ConstructBlueprint
        bp_data = {'id': data.get('blueprint_id'),
                   'processes': data.get('processes'),
                   'model_type': data.get('model_type'),
                   'project_id': data.get('project_id')}
        self.blueprint = Blueprint.from_data(bp_data)

    @classmethod
    def _job_type(cls):
        return enums.JOB_TYPE.MODEL

    @classmethod
    def from_job(cls, job):
        """ Transforms a generic Job into a ModelJob

        Parameters
        ----------
        job: Job
            A generic job representing a ModelJob

        Returns
        -------
        model_job: ModelJob
            A fully populated ModelJob with all the details of the job

        Raises
        ------
        ValueError:
            If the generic Job was not a model job, e.g. job_type != JOB_TYPE.MODEL
        """
        return super(ModelJob, cls).from_job(job)

    @classmethod
    def get(cls, project_id, model_job_id):
        """
        Fetches one ModelJob. If the job finished, raises PendingJobFinished
        exception.

        Parameters
        ----------
        project_id : str
            The identifier of the project the model belongs to
        model_job_id : str
            The identifier of the model_job

        Returns
        -------
        model_job : ModelJob
            The pending ModelJob

        Raises
        ------
        PendingJobFinished
            If the job being queried already finished, and the server is
            re-routing to the finished model.
        AsyncFailureError
            Querying this resource gave a status code other than 200 or 303
        """
        return super(ModelJob, cls).get(project_id, model_job_id)

    @classmethod
    def _job_path(cls, project_id, job_id):
        return 'projects/{}/modelJobs/{}/'.format(project_id, job_id)

    @classmethod
    def get_model(cls, project_id, model_job_id):
        """
        Fetches a finished model from the job used to create it.

        Parameters
        ----------
        project_id : str
            The identifier of the project the model belongs to
        model_job_id : str
            The identifier of the model_job

        Returns
        -------
        model : Model
            The finished model

        Raises
        ------
        JobNotFinished
            If the job has not finished yet
        AsyncFailureError
            Querying the model_job in question gave a status code other than 200 or
            303
        """
        # Note: Don't need this in 3.0. (Use `Job.get_result` method instead.)
        from . import Model
        url = cls._job_path(project_id, model_job_id)
        response = cls._client.get(url, allow_redirects=False)
        if response.status_code == 200:
            data = response.json()
            raise errors.JobNotFinished('Pending job status is {}'.format(data['status']))
        elif response.status_code == 303:
            location = response.headers['Location']
            return Model.from_location(location)
        else:
            e_msg = 'Server unexpectedly returned status code {}'
            raise errors.AsyncFailureError(
                e_msg.format(response.status_code))



import json
import os
import tempfile
import warnings

import mock
import pytest
import responses

from datarobot import (Project, Model, PrimeFile, PrimeModel, Ruleset, enums,
                       BlenderModel, FrozenModel, PredictionExplanationsInitialization,
                       PredictionExplanations, ReasonCodesInitialization, ReasonCodes,
                       RatingTable, RatingTableModel)
from datarobot.models.prediction_explanations import PredictionExplanationsPage
from datarobot.models.reason_codes import ReasonCodesPage
from datarobot.rest import RESTClientObject
from datarobot.utils import from_api
from datarobot.client import get_client, set_client


# This filter causes the tests to fail if any uncaught warnings leak out
warnings.simplefilter('error')
# some of our dependencies trigger DeprecationWarning's on Python 3.7 (internally), which we would
# normally treat as exceptions per the rule above; ignore them instead so that tests pass on 3.7
warnings.filterwarnings(
    'ignore',
    category=DeprecationWarning,
    message=r'.*',
    module=r'(requests)|(responses)|(yaml)',
)


@pytest.yield_fixture
def temporary_file():
    new_file = tempfile.NamedTemporaryFile(delete=False)
    new_file.close()
    yield new_file.name
    os.remove(new_file.name)


@pytest.yield_fixture
def mock_async_time():
    patch = mock.patch('datarobot.utils.waiters.time')
    yield patch.start()
    patch.stop()


@pytest.fixture
def unicode_string():
    return u'\u3053\u3093\u306b\u3061\u306f'


@pytest.fixture
def unittest_endpoint():
    return 'https://host_name.com'


@pytest.yield_fixture(scope='function')
def known_warning():
    """
    A context that will not let any warnings out. This will allow us to
    test for known deprecations and the like while still making sure the rest of the tested
    code does not emit any warnings

    This fixture asserts that a warning was raised while it was used, so users of this
    fixture don't need to do so themselves
    """
    filters = warnings.filters[:]
    warnings.resetwarnings()
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')  # record _all_ warnings, not just first instance
        yield w
    assert w, 'Expected a warning but did not find one'
    warnings.filters = filters


@pytest.yield_fixture(scope='function')
def client():
    """A mocked client

    The DataRobot package is built around acquiring the GlobalClient, which this sets
    to point to `https://host_name.com`.

    Most often you only need the effect, not the client. In this case you can use
    the pytest.mark.usefixtures decorator to make sure this patch is takes place
    for your test
    """
    c = RESTClientObject(auth='t-token', endpoint='https://host_name.com')
    set_client(c)
    yield get_client()
    set_client(None)


@pytest.fixture
def project_id():
    """
    A project id that matches the objects in the fixtures

    Returns
    -------
    project_id : str
        The id of the project used across the fixtures
    """
    return '556cdfbb100d2b0e88585195'


@pytest.fixture
def project_collection_url():
    return 'https://host_name.com/projects/'


@pytest.fixture
def project_url(project_id):
    return 'https://host_name.com/projects/{}/'.format(project_id)


@pytest.fixture
def project_aim_url(project_url):
    return '{}{}/'.format(project_url, 'aim')


@pytest.fixture
def project_without_target_json():
    """The JSON of one project

    This data in this project has been uploaded and analyzed, but the target
    has not been set
    """
    return """
    {
    "id": "556cdfbb100d2b0e88585195",
    "projectName": "A Project Name",
    "fileName": "test_data.csv",
    "stage": "aim",
    "autopilotMode": null,
    "created": "2016-07-26T02:29:58.546312Z",
    "target": null,
    "metric": null,
    "partition": {
      "datetimeCol": null,
      "cvMethod": null,
      "validationPct": null,
      "reps": null,
      "cvHoldoutLevel": null,
      "holdoutLevel": null,
      "userPartitionCol": null,
      "validationType": null,
      "trainingLevel": null,
      "partitionKeyCols": null,
      "holdoutPct": null,
      "validationLevel": null
    },
    "recommender": {
      "recommenderItemId": null,
      "isRecommender": null,
      "recommenderUserId": null
    },
    "advancedOptions": {
      "blueprintThreshold": null,
      "responseCap": null,
      "seed": null,
      "weights": null,
      "smartDownsampled": null,
      "majorityDownsamplingRate": null,
      "offset": null,
      "exposure": null
    },
    "positiveClass": null,
    "maxTrainPct": null,
    "holdoutUnlocked": false,
    "targetType": null
  }
"""


@pytest.fixture
def project_without_target_data(project_without_target_json):
    data = json.loads(project_without_target_json)
    return from_api(data)


@pytest.fixture
def project_with_target_json():
    return """{
        "id": "556cdfbb100d2b0e88585195",
        "projectName": "A Project Name",
        "fileName": "data.csv",
        "stage": "modeling",
        "autopilotMode": 0,
        "created": "2016-07-26T02:29:58.546312Z",
        "target": "target_name",
        "metric": "LogLoss",
        "partition": {
            "datetimeCol": null,
            "cvMethod": "stratified",
            "validationPct": null,
            "reps": 5,
            "cvHoldoutLevel": null,
            "holdoutLevel": null,
            "userPartitionCol": null,
            "validationType": "CV",
            "trainingLevel": null,
            "partitionKeyCols": null,
            "holdoutPct": 19.99563,
            "validationLevel": null
        },
        "recommender": {
            "recommenderItemId": null,
            "isRecommender": false,
            "recommenderUserId": null
        },
        "advancedOptions": {
            "blueprintThreshold": null,
            "responseCap": false,
            "seed": null,
            "weights": null,
            "smartDownsampled": false,
            "majorityDownsamplingRate": null,
            "offset": null,
            "exposure": null
        },
        "positiveClass": 1,
        "maxTrainPct": 64.0035,
        "holdoutUnlocked": false,
        "targetType": "Binary"
    }
"""


@pytest.fixture
def project_with_target_data(project_with_target_json):
    data = json.loads(project_with_target_json)
    return from_api(data)


@pytest.fixture
def smart_sampled_project_server_data(project_with_target_json):
    base_data = json.loads(project_with_target_json)
    return dict(base_data, smartDownsampled=True, majorityDownsamplingRate=50.5)


@pytest.fixture
def project(project_with_target_data):
    return Project.from_data(project_with_target_data)


@pytest.fixture
def smart_sampled_project(smart_sampled_project_data):
    return Project.from_data(smart_sampled_project_data)


@pytest.fixture
def project_without_target(project_without_target_data):
    return Project.from_data(project_without_target_data)


@pytest.fixture
def async_failure_json():
    return """
    {
        "status": "ERROR",
        "message": "",
        "code": 0,
        "created": "2016-07-22T12:00:00.123456Z"
    }
    """


@pytest.fixture
def async_aborted_json():
    return """
    {
        "status": "ABORTED",
        "message": "",
        "code": 0,
        "created": "2016-07-22T12:00:00.123456Z"
    }
    """


@pytest.fixture
def async_running_json():
    return """
    {
        "status": "RUNNING",
        "message": "",
        "code": 0,
        "created": "2016-07-22T12:00:00.123456Z"
    }
    """


@pytest.fixture
def featurelist_server_data(project_id):
    return {
        'id': '5223deadbeefdeadbeef9999',
        'name': 'Some Features',
        'features': ['One Fish', 'Two Fish', 'Red Fish', 'Blue Fish'],
        'created': '2018-07-02T18:49:38.019000Z',
        'isUserCreated': True,
        'numModels': 5,
        'projectId': project_id,
        'description': 'Some String'
    }


@pytest.fixture
def model_id():
    """
    The id of the model used in the fixtures

    Returns
    -------
    model_id : str
        The id of the model used in the fixtures
    """
    return '556ce973100d2b6e51ca9657'


@pytest.fixture
def model_json():
    return """
{
    "featurelistId": "57993241bc92b715ed0239ee",
    "processes": [
      "One",
      "Two",
      "Three"
    ],
    "featurelistName": "Informative Features",
    "projectId": "556cdfbb100d2b0e88585195",
    "samplePct": 64,
    "modelType": "Gradient Boosted Trees Classifier",
    "metrics": {
      "AUC": {
        "holdout": null,
        "validation": 0.73376,
        "crossValidation": null
      },
      "Rate@Top5%": {
        "holdout": null,
        "validation": 0.44218,
        "crossValidation": null
      },
      "Rate@TopTenth%": {
        "holdout": null,
        "validation": 1,
        "crossValidation": null
      },
      "RMSE": {
        "holdout": null,
        "validation": 0.27966,
        "crossValidation": null
      },
      "LogLoss": {
        "holdout": null,
        "validation": 0.2805,
        "crossValidation": null
      },
      "FVE Binomial": {
        "holdout": null,
        "validation": 0.12331,
        "crossValidation": null
      },
      "Gini Norm": {
        "holdout": null,
        "validation": 0.46752,
        "crossValidation": null
      },
      "Rate@Top10%": {
        "holdout": null,
        "validation": 0.34812,
        "crossValidation": null
      }
    },
    "modelCategory": "model",
    "isFrozen": false,
    "blueprintId": "de628edee06f2b23218767a245e45ae1",
    "id": "556ce973100d2b6e51ca9657",
    "supportsMonotonicConstraints": true,
    "monotonicIncreasingFeaturelistId": "5ae04c0d962d7410683073cb",
    "monotonicDecreasingFeaturelistId": "5ae04c0f962d7410683073cc",
    "isStarred": false,
    "predictionThreshold": 0.6,
    "predictionThresholdReadOnly": true
  }
    """


@pytest.fixture
def model_deployment_id():
    return "5a7d96bd962d745ea62459fb"


@pytest.fixture
def model_deployment_json():
    return """
{
    "status": "active",
    "trendTimeWindow": "1d",
    "prevRequestCount": 0,
    "description": "test",
    "predictionEndpoint": "fake-5a7d9559962d745ea62459c7",
    "organizationId": "5a7d96a3962d745ea62459f1",
    "requestRates": [0, 24, 24, 24, 23, 0, 0, 0],
    "createdAt": "2018-02-09 12:40:29.730000",
    "label": "Auto-Tuned Word N-Gram Text Modeler using token occurrences - diag_1_desc",
    "project": {
        "id": "5a7d69ad962d743f559d1a97",
        "projectName": "MMM",
        "fileName":
        "10k_diabetes-200.tsv",
        "stage":
        "modeling",
        "autopilotMode": 0,
        "created": "2018-02-09T09:28:19.928403Z",
        "target": "readmitted",
        "metric": "LogLoss",
        "partition": {
            "datetimeCol": null,
            "cvHoldoutLevel": null,
            "validationPct": null,
            "reps": 5,
            "cvMethod":
            "stratified",
            "holdoutLevel": null,
            "userPartitionCol": null,
            "validationType": "CV",
            "trainingLevel": null,
            "partitionKeyCols": null,
            "holdoutPct": 20.0,
            "validationLevel": null
        },
        "recommender": {
            "recommenderItemId": null,
            "isRecommender": null,
            "recommenderUserId": null
        },
        "advancedOptions": {
            "scaleoutModelingMode": "disabled",
            "responseCap": false,
            "downsampledMinorityRows": null,
            "smartDownsampled": false,
            "blueprintThreshold": 3,
            "seed": null,
            "weights": null,
            "downsampledMajorityRows": null,
            "majorityDownsamplingRate": null
        },
        "positiveClass": 1.0,
        "maxTrainPct": 64.0,
        "maxTrainRows": 128,
        "scaleoutMaxTrainPct": 64.0,
        "scaleoutMaxTrainRows": 128,
        "holdoutUnlocked": true,
         "targetType": "Binary"
    },
    "instance": null,
    "serviceHealthMessages": [
        {"msgId": "NO_GOOD_REQUESTS",
         "message": "No successful predictions in 24 hours",
         "level": "passing"}
    ],
    "model": {
        "uid": "5a2e9f83c15fa0be8d2b2c74",
        "id": "5a7d69dd962d745e1b69535b",
        "modelType": "Auto-Tuned Word N-Gram Text Modeler using token occurrences - diag_1_desc",
        "icons": [0]
    },
    "user": {
        "username": "admin@datarobot.com",
        "lastName": "Test",
        "id": "5a2e9f83c15fa0be8d2b2c74",
        "firstName": "Test"
    },
    "relativeRequestsTrend": 5,
    "updatedAt": "2018-02-12 12:16:39.163000",
    "serviceHealth": "passing",
    "recentRequestCount": 10,
    "type": "dedicated",
    "id": "5a7d96bd962d745ea62459fb",
    "deployed": true
}
    """


@pytest.fixture
def blender_json(model_json):
    data = json.loads(model_json)
    data['modelIds'] = ['556ce973100d2b6e51ca9656', '556ce973100d2b6e51ca9655']
    data['blenderMethod'] = 'AVG'
    return json.dumps(data)


@pytest.fixture
def blenders_list_response_json(blender_json):
    data = json.loads(blender_json)
    return json.dumps({'data': [data, data]})


@pytest.fixture
def frozen_json(model_json):
    data = json.loads(model_json)
    data['parentModelId'] = '556ce973100e2b6e51ca8657'
    return json.dumps(data)


@pytest.fixture
def frozen_models_list_response(frozen_json):
    data = json.loads(frozen_json)
    return {'data': [data, data], 'count': 2, 'prev': None, 'next': None}


@pytest.fixture
def model_data(model_json):
    return from_api(json.loads(model_json))


@pytest.fixture
def model_deployment_data(model_deployment_json):
    return from_api(json.loads(model_deployment_json))


@pytest.fixture
def datetime_model_data(model_data):
    modified_model_data = dict(model_data)
    modified_model_data.pop('sample_pct')
    modified_model_data['training_start_date'] = '2015-12-10T19:00:00.000000Z'
    modified_model_data['training_end_date'] = '2016-12-10T19:00:00.000000Z'
    return modified_model_data


@pytest.fixture
def one_model(model_data):
    return Model.from_data(model_data)


@pytest.fixture
def blender_data(blender_json):
    return from_api(json.loads(blender_json))


@pytest.fixture
def one_blender(blender_data):
    return BlenderModel.from_data(blender_data)


@pytest.fixture
def frozen_data(frozen_json):
    return from_api(json.loads(frozen_json))


@pytest.fixture
def one_frozen(frozen_data):
    return FrozenModel.from_data(frozen_data)


@pytest.fixture
def prime_model_id():
    return '57aa68e1ccf94e1ce3197743'


@pytest.fixture
def prime_model_json():
    return """
    {
    "featurelistId": "57aa1c46ccf94e5bb073841b",
    "processes": [],
    "featurelistName": "Informative Features",
    "projectId": "556cdfbb100d2b0e88585195",
    "samplePct": 63.863,
    "modelType": "DataRobot Prime",
    "metrics": {
      "AUC": {
        "holdout": null,
        "validation": 0.8559,
        "crossValidation": null
      },
      "Rate@Top5%": {
        "holdout": null,
        "validation": 1,
        "crossValidation": null
      },
      "Rate@TopTenth%": {
        "holdout": null,
        "validation": 1,
        "crossValidation": null
      },
      "RMSE": {
        "holdout": null,
        "validation": 0.37973,
        "crossValidation": null
      },
      "LogLoss": {
        "holdout": null,
        "validation": 0.41848,
        "crossValidation": null
      },
      "FVE Binomial": {
        "holdout": null,
        "validation": 0.32202,
        "crossValidation": null
      },
      "Gini Norm": {
        "holdout": null,
        "validation": 0.7118,
        "crossValidation": null
      },
      "Rate@Top10%": {
        "holdout": null,
        "validation": 0.66667,
        "crossValidation": null
      }
    },
    "modelCategory": "prime",
    "blueprintId": "bcfb575932da72a92d01837a6c42a36f5cc56cbdab7d92f43b88e114179f2942",
    "id": "57aa68e1ccf94e1ce3197743",
    "rulesetId": 3,
    "score": 0.41847989771503824,
    "parentModelId": "556ce973100d2b6e51ca9657",
    "ruleCount": 323
    }
    """


@pytest.fixture
def prime_model_server_data(prime_model_json):
    return json.loads(prime_model_json)


@pytest.fixture
def prime_data(prime_model_server_data):
    return from_api(prime_model_server_data)


@pytest.fixture
def prime_model(prime_model_server_data):
    return PrimeModel.from_server_data(prime_model_server_data)


@pytest.fixture
def ruleset_with_model_json():
    return """
    {
    "projectId": "556cdfbb100d2b0e88585195",
    "rulesetId": 3,
    "score": 0.41847989771503824,
    "parentModelId": "556ce973100d2b6e51ca9657",
    "ruleCount": 323,
    "modelId": "57aa68e1ccf94e1ce3197743"
    }
    """


@pytest.fixture
def ruleset_without_model_json():
    return """
    {
    "projectId": "556cdfbb100d2b0e88585195",
    "rulesetId": 2,
    "score": 0.428702,
    "parentModelId": "556ce973100d2b6e51ca9657",
    "ruleCount": 161,
    "modelId": null
    }
    """


@pytest.fixture
def ruleset_with_model_server_data(ruleset_with_model_json):
    return json.loads(ruleset_with_model_json)


@pytest.fixture
def ruleset_without_model_server_data(ruleset_without_model_json):
    return json.loads(ruleset_without_model_json)


@pytest.fixture
def ruleset_with_model(ruleset_with_model_server_data):
    return Ruleset.from_server_data(ruleset_with_model_server_data)


@pytest.fixture
def ruleset_without_model(ruleset_without_model_server_data):
    return Ruleset.from_server_data(ruleset_without_model_server_data)


@pytest.fixture
def prime_file_json():
    return """
    {
    "id": "57fa1c41ccf94e59a9024e87",
    "projectId": "556cdfbb100d2b0e88585195",
    "parentModelId": "556ce973100d2b6e51ca9657",
    "modelId": "57aa68e1ccf94e1ce3197743",
    "rulesetId": 3,
    "language": "Python",
    "isValid": true
    }
    """


@pytest.fixture
def prime_file_server_data(prime_file_json):
    return json.loads(prime_file_json)


@pytest.fixture
def prime_file(prime_file_server_data):
    return PrimeFile.from_server_data(prime_file_server_data)


@pytest.fixture
def rating_table_model_server_data(model_json):
    data = json.loads(model_json)
    data['ratingTableId'] = "5984af32100d2b2d9f10fbe0"
    return data


@pytest.fixture
def rating_table_model_json(rating_table_model_server_data):
    return json.dumps(rating_table_model_server_data)


@pytest.fixture
def rating_table_model(rating_table_model_server_data):
    return RatingTableModel.from_server_data(rating_table_model_server_data)


@pytest.fixture
def rating_table_model_url(project_collection_url, rating_table_model_server_data):
    d = rating_table_model_server_data
    return '{}{}/ratingTableModels/{}/'.format(project_collection_url, d['projectId'], d['id'])


@pytest.fixture
def job_id():
    return '13'


@pytest.fixture
def job_url(project_id, job_id):
    return 'https://host_name.com/projects/{}/jobs/{}/'.format(project_id, job_id)


@pytest.fixture
def base_job_server_data(job_id, project_id, job_url):
    return {
        'status': None,
        'url': job_url,
        'id': job_id,
        'jobType': None,
        'isBlocked': False,
        'projectId': project_id
    }


@pytest.fixture
def base_job_running_server_data(base_job_server_data):
    return dict(base_job_server_data, status=enums.QUEUE_STATUS.INPROGRESS)


@pytest.fixture
def base_job_completed_server_data(base_job_server_data):
    return dict(base_job_server_data, status=enums.QUEUE_STATUS.COMPLETED)


@pytest.fixture
def prime_model_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_MODEL)


@pytest.fixture
def prime_model_job_completed_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_MODEL)


@pytest.fixture
def prime_model_job_creation_response(project_url, job_url):
    responses.add(responses.POST, '{}primeModels/'.format(project_url), body='', status=202,
                  adding_headers={'Location': job_url})


@pytest.fixture
def prime_model_job_completed_response(prime_model_job_completed_server_data,
                                       job_url, project_url, prime_model_id,
                                       prime_model_server_data):
    prime_model_url = '{}primeModels/{}/'.format(project_url, prime_model_id)
    responses.add(responses.GET, job_url, body=json.dumps(prime_model_job_completed_server_data),
                  status=303, adding_headers={'Location': prime_model_url},
                  content_type='application/json')
    responses.add(responses.GET, prime_model_url, body=json.dumps(prime_model_server_data),
                  content_type='application/json')


@pytest.fixture
def frozen_model_job_completed_server_data(base_job_completed_server_data, one_model):
    return dict(base_job_completed_server_data,
                jobType=enums.JOB_TYPE.MODEL,
                modelType=one_model.model_type,
                processes=one_model.processes,
                blueprintId=one_model.blueprint_id)


@pytest.fixture
def rci_json():
    """ ReasonCodesInitialization GET json
    """
    return """{
      "projectId": "556cdfbb100d2b0e88585195",
      "reasonCodesSample": [
        {
          "reasonCodes": [
            {
              "featureValue": 57000,
              "label": "YearMade",
              "strength": 115.59591064453161,
              "feature": "SalePrice",
              "qualitativeStrength": "+++"
            },
            {
              "featureValue": "Wheel Loader - 250.0 to 275.0 Horsepower",
              "label": "YearMade",
              "strength": 60.62447509765593,
              "feature": "fiProductClassDesc",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "",
              "label": "YearMade",
              "strength": 54.968029785156205,
              "feature": "Stick_Length",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 1.0,
              "label": "YearMade",
              "strength": 51.503918457031205,
              "feature": "auctioneerID",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 1763636,
              "label": "YearMade",
              "strength": 45.668481445312636,
              "feature": "SalesID",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "-2",
              "label": "YearMade",
              "strength": 35.2052001953125,
              "feature": "fiModelSeries",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "Utah",
              "label": "YearMade",
              "strength": 25.33120117187491,
              "feature": "state",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "WA450",
              "label": "YearMade",
              "strength": 21.51729736328116,
              "feature": "fiBaseModel",
              "qualitativeStrength": "+"
            },
            {
              "featureValue": 383693,
              "label": "YearMade",
              "strength": 17.533557128906068,
              "feature": "MachineID",
              "qualitativeStrength": "+"
            },
            {
              "featureValue": "",
              "label": "YearMade",
              "strength": 12.142431640625091,
              "feature": "fiSecondaryDesc",
              "qualitativeStrength": "+"
            }
          ],
          "prediction": 2059.988037109375,
          "rowId": 7459,
          "predictionValues": [{"value": 2059.988037109375, "label": "YearMade"}]
        },
        {
          "reasonCodes": [
            {
              "featureValue": "50DZTS",
              "label": "YearMade",
              "strength": 86.08530273437486,
              "feature": "fiModelDesc",
              "qualitativeStrength": "+++"
            },
            {
              "featureValue": 2063.0,
              "label": "YearMade",
              "strength": 49.44567871093727,
              "feature": "MachineHoursCurrentMeter",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "Medium",
              "label": "YearMade",
              "strength": 43.51775771484381,
              "feature": "UsageBand",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "Hydraulic Excavator, Track - 4.0 to 5.0 Metric Tons",
              "label": "YearMade",
              "strength": 33.40051269531273,
              "feature": "fiProductClassDesc",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 1250726400,
              "label": "YearMade",
              "strength": 30.625183105468295,
              "feature": "saledate",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "",
              "label": "YearMade",
              "strength": 26.95623779296875,
              "feature": "Stick_Length",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 21500,
              "label": "YearMade",
              "strength": 25.61611328124968,
              "feature": "SalePrice",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 22117,
              "label": "YearMade",
              "strength": 25.539965820312545,
              "feature": "ModelID",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "D",
              "label": "YearMade",
              "strength": 21.061962890624955,
              "feature": "fiSecondaryDesc",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "50",
              "label": "YearMade",
              "strength": 20.641674804687682,
              "feature": "fiBaseModel",
              "qualitativeStrength": "++"
            }
          ],
          "prediction": 2056.3349609375,
          "rowId": 6934,
          "predictionValues": [{"value": 2056.3349609375, "label": "YearMade"}]
        }
      ],
      "modelId": "578e59a41ced2e5a9eb18965"
    }"""


@pytest.fixture
def rci_data(rci_json):
    return from_api(json.loads(rci_json))


@pytest.fixture
def one_rci(rci_data):
    return ReasonCodesInitialization.from_data(rci_data)


@pytest.fixture
def rc_list_json():
    """List of reason codes"""
    return """{
      "count": 3,
      "next": null,
      "data": [
        {
          "finishTime": 1482414443.292265,
          "numColumns": 50,
          "reasonCodesLocation": "https://host_name.com/projects/5..4/reasonCodes/5..7/",
          "thresholdHigh": 0.835627812,
          "projectId": "578e565a218ef11f13fd2f64",
          "thresholdLow": 0.1,
          "maxCodes": 2,
          "id": "585bd862dd56a7482bdcb567",
          "datasetId": "585bbd1dc522f21821107d6e",
          "modelId": "578e59a41ced2e5a9eb18965"
        },
        {
          "finishTime": 1482414817.200652,
          "numColumns": 50,
          "reasonCodesLocation": "https://host_name.com/projects/5..4/reasonCodes/5..6/",
          "thresholdHigh": 0.835627812,
          "projectId": "578e565a218ef11f13fd2f64",
          "thresholdLow": 0.1,
          "maxCodes": 3,
          "id": "585bda05dd56a7483781c4b6",
          "datasetId": "585bbd1dc522f21821107d6e",
          "modelId": "578e59a41ced2e5a9eb18965"
        },
        {
          "finishTime": 1482418494.81697,
          "numColumns": 50,
          "reasonCodesLocation": "https://host_name.com/projects/5..4/reasonCodes/5..a/",
          "thresholdHigh": 0.835627812,
          "projectId": "578e565a218ef11f13fd2f64",
          "thresholdLow": 0.1,
          "maxCodes": 3,
          "id": "585be85cdd56a7482bdcb56a",
          "datasetId": "585bbd1dc522f21821107d6e",
          "modelId": "578e59a41ced2e5a9eb18965"
        }
      ],
      "previous": null
    }"""


@pytest.fixture
def rc_json():
    """ ReasonCodes GET json
    """
    return """{
      "finishTime": 1482424475.33823,
      "numColumns": 50,
      "reasonCodesLocation": "https://host_name.com/projects/578e...2f64/reasonCodes/585b...9e0f/",
      "thresholdHigh": 2010.0490966796874,
      "projectId": "556cdfbb100d2b0e88585195",
      "thresholdLow": 1706.8041999316963,
      "maxCodes": 3,
      "id": "585ba071dd56a72ec7109e0f",
      "datasetId": "585b9d71c522f20d38107d6e",
      "modelId": "578e59a41ced2e5a9eb18960"
    }"""


@pytest.fixture
def rc_data(rc_json):
    return from_api(json.loads(rc_json))


@pytest.fixture
def one_rc(rc_data):
    return ReasonCodes.from_data(rc_data)


@pytest.fixture
def rcp_json():
    """ ReasonCodesPage GET json
    """
    return """{
      "count": 6,
      "reasonCodesRecordLocation": "https://host_name.com/projects/5..4/reasonCodesRecords/5..7/",
      "next": null,
      "data": [
        {
          "reasonCodes": [
            {
              "featureValue": 66000,
              "label": "YearMade",
              "strength": 113.18563232421889,
              "feature": "SalePrice",
              "qualitativeStrength": "+++"
            },
            {
              "featureValue": 1139246,
              "label": "YearMade",
              "strength": -84.37227783203139,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            }
          ],
          "prediction": 1843.3448486328125,
          "rowId": 0,
          "predictionValues": [
            {
              "value": 1843.3448486328125,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 121,
              "label": "YearMade",
              "strength": -140.4251185424805,
              "feature": "datasource",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 57000,
              "label": "YearMade",
              "strength": 107.4373168945308,
              "feature": "SalePrice",
              "qualitativeStrength": "++"
            }
          ],
          "prediction": 1730.043701171875,
          "rowId": 1,
          "predictionValues": [
            {
              "value": 1730.043701171875,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 1139249,
              "label": "YearMade",
              "strength": -201.79853515624995,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 10000,
              "label": "YearMade",
              "strength": -144.66733398437532,
              "feature": "SalePrice",
              "qualitativeStrength": "---"
            }
          ],
          "prediction": 1751.1256103515625,
          "rowId": 2,
          "predictionValues": [
            {
              "value": 1751.1256103515625,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 1139251,
              "label": "YearMade",
              "strength": -222.99357910156255,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 1305763200,
              "label": "YearMade",
              "strength": -107.39191894531268,
              "feature": "saledate",
              "qualitativeStrength": "--"
            }
          ],
          "prediction": 1663.2529296875,
          "rowId": 3,
          "predictionValues": [
            {
              "value": 1663.2529296875,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 1139253,
              "label": "YearMade",
              "strength": -209.14246826171893,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 11000,
              "label": "YearMade",
              "strength": -67.03535156250018,
              "feature": "SalePrice",
              "qualitativeStrength": "--"
            }
          ],
          "prediction": 1724.86328125,
          "rowId": 4,
          "predictionValues": [
            {
              "value": 1724.86328125,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 1139255,
              "label": "YearMade",
              "strength": -81.31390380859398,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 26500,
              "label": "YearMade",
              "strength": 52.808850097656205,
              "feature": "SalePrice",
              "qualitativeStrength": "++"
            }
          ],
          "prediction": 1881.0067138671875,
          "rowId": 5,
          "predictionValues": [
            {
              "value": 1881.0067138671875,
              "label": "YearMade"
            }
          ]
        }
      ],
      "id": "585bd862dd56a7482bdcb567",
      "previous": null
    }"""


@pytest.fixture
def rcp_data(rcp_json):
    return from_api(json.loads(rcp_json))


@pytest.fixture
def rcp_data_with_adjusted_predictions(rcp_data):
    rcp_data_with_ap = rcp_data.copy()
    rcp_data_with_ap['adjustment_method'] = 'exposureNormalized'
    for record in rcp_data_with_ap['data']:
        # Trivial case, all exposure values set to 1
        record['adjusted_prediction'] = record['prediction']
        record['adjusted_prediction_values'] = record['prediction_values']
    return rcp_data_with_ap


@pytest.fixture
def rcp_json_classification():
    return """{
      "count": 4,
      "reasonCodesRecordLocation": "https://host_name.com/projects/5...2/reasonCodesRecords/5...c/",
      "next": null,
      "data": [
        {
          "reasonCodes": [],
          "prediction": 0,
          "rowId": 0,
          "predictionValues": [
            {
              "value": 0.2893982637455831,
              "label": 1
            },
            {
              "value": 0.7106017362544169,
              "label": 0
            }
          ]
        },
        {
          "reasonCodes": [],
          "prediction": 0,
          "rowId": 1,
          "predictionValues": [
            {
              "value": 0.29246609348399255,
              "label": 1
            },
            {
              "value": 0.7075339065160074,
              "label": 0
            }
          ]
        },
        {
          "reasonCodes": [],
          "prediction": 0,
          "rowId": 2,
          "predictionValues": [
            {
              "value": 0.3840762929836601,
              "label": 1
            },
            {
              "value": 0.6159237070163399,
              "label": 0
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 3,
              "strength": -0.3712657301791762,
              "feature": "number_diagnoses",
              "qualitativeStrength": "--",
              "label": 1
            },
            {
              "featureValue": "UN",
              "strength": -0.3112189393477769,
              "feature": "payer_code",
              "qualitativeStrength": "--",
              "label": 1
            },
            {
              "featureValue": "41",
              "strength": -0.2607262089753333,
              "feature": "diag_2",
              "qualitativeStrength": "--",
              "label": 1
            }
          ],
          "prediction": 0,
          "rowId": 3,
          "predictionValues": [
            {
              "value": 0.12064479565777504,
              "label": 1
            },
            {
              "value": 0.8793552043422249,
              "label": 0
            }
          ]
        }
      ],
      "id": "58d1ef9fc80891088925890c",
      "previous": null
    }"""


@pytest.fixture
def rcp_data_classification(rcp_json_classification):
    return from_api(json.loads(rcp_json_classification))


@pytest.fixture
def one_rcp(rcp_data):
    with mock.patch('datarobot.models.reason_codes.t.URL.check',
                    side_effect=(lambda url, context=None: url)):
        return ReasonCodesPage.from_data(rcp_data)


@pytest.fixture
def pei_json():
    """ PredictionExplanationsInitialization GET json
    """
    return """{
      "projectId": "556cdfbb100d2b0e88585195",
      "predictionExplanationsSample": [
        {
          "predictionExplanations": [
            {
              "featureValue": 57000,
              "label": "YearMade",
              "strength": 115.59591064453161,
              "feature": "SalePrice",
              "qualitativeStrength": "+++"
            },
            {
              "featureValue": "Wheel Loader - 250.0 to 275.0 Horsepower",
              "label": "YearMade",
              "strength": 60.62447509765593,
              "feature": "fiProductClassDesc",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "",
              "label": "YearMade",
              "strength": 54.968029785156205,
              "feature": "Stick_Length",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 1.0,
              "label": "YearMade",
              "strength": 51.503918457031205,
              "feature": "auctioneerID",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 1763636,
              "label": "YearMade",
              "strength": 45.668481445312636,
              "feature": "SalesID",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "-2",
              "label": "YearMade",
              "strength": 35.2052001953125,
              "feature": "fiModelSeries",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "Utah",
              "label": "YearMade",
              "strength": 25.33120117187491,
              "feature": "state",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "WA450",
              "label": "YearMade",
              "strength": 21.51729736328116,
              "feature": "fiBaseModel",
              "qualitativeStrength": "+"
            },
            {
              "featureValue": 383693,
              "label": "YearMade",
              "strength": 17.533557128906068,
              "feature": "MachineID",
              "qualitativeStrength": "+"
            },
            {
              "featureValue": "",
              "label": "YearMade",
              "strength": 12.142431640625091,
              "feature": "fiSecondaryDesc",
              "qualitativeStrength": "+"
            }
          ],
          "prediction": 2059.988037109375,
          "rowId": 7459,
          "predictionValues": [{"value": 2059.988037109375, "label": "YearMade"}]
        },
        {
          "predictionExplanations": [
            {
              "featureValue": "50DZTS",
              "label": "YearMade",
              "strength": 86.08530273437486,
              "feature": "fiModelDesc",
              "qualitativeStrength": "+++"
            },
            {
              "featureValue": 2063.0,
              "label": "YearMade",
              "strength": 49.44567871093727,
              "feature": "MachineHoursCurrentMeter",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "Medium",
              "label": "YearMade",
              "strength": 43.51775771484381,
              "feature": "UsageBand",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "Hydraulic Excavator, Track - 4.0 to 5.0 Metric Tons",
              "label": "YearMade",
              "strength": 33.40051269531273,
              "feature": "fiProductClassDesc",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 1250726400,
              "label": "YearMade",
              "strength": 30.625183105468295,
              "feature": "saledate",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "",
              "label": "YearMade",
              "strength": 26.95623779296875,
              "feature": "Stick_Length",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 21500,
              "label": "YearMade",
              "strength": 25.61611328124968,
              "feature": "SalePrice",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 22117,
              "label": "YearMade",
              "strength": 25.539965820312545,
              "feature": "ModelID",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "D",
              "label": "YearMade",
              "strength": 21.061962890624955,
              "feature": "fiSecondaryDesc",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "50",
              "label": "YearMade",
              "strength": 20.641674804687682,
              "feature": "fiBaseModel",
              "qualitativeStrength": "++"
            }
          ],
          "prediction": 2056.3349609375,
          "rowId": 6934,
          "predictionValues": [{"value": 2056.3349609375, "label": "YearMade"}]
        }
      ],
      "modelId": "578e59a41ced2e5a9eb18965"
    }"""


@pytest.fixture
def pei_data(pei_json):
    return from_api(json.loads(pei_json))


@pytest.fixture
def one_pei(pei_data):
    return PredictionExplanationsInitialization.from_data(pei_data)


@pytest.fixture
def pe_list_json():
    """List of reason codes"""
    return """{
      "count": 3,
      "next": null,
      "data": [
        {
          "finishTime": 1482414443.292265,
          "numColumns": 50,
          "predictionExplanationsLocation":
            "https://host_name.com/projects/5..4/predictionExplanations/5..7/",
          "thresholdHigh": 0.835627812,
          "projectId": "578e565a218ef11f13fd2f64",
          "thresholdLow": 0.1,
          "maxExplanations": 2,
          "id": "585bd862dd56a7482bdcb567",
          "datasetId": "585bbd1dc522f21821107d6e",
          "modelId": "578e59a41ced2e5a9eb18965"
        },
        {
          "finishTime": 1482414817.200652,
          "numColumns": 50,
          "predictionExplanationsLocation":
            "https://host_name.com/projects/5..4/predictionExplanations/5..6/",
          "thresholdHigh": 0.835627812,
          "projectId": "578e565a218ef11f13fd2f64",
          "thresholdLow": 0.1,
          "maxExplanations": 3,
          "id": "585bda05dd56a7483781c4b6",
          "datasetId": "585bbd1dc522f21821107d6e",
          "modelId": "578e59a41ced2e5a9eb18965"
        },
        {
          "finishTime": 1482418494.81697,
          "numColumns": 50,
          "predictionExplanationsLocation":
            "https://host_name.com/projects/5..4/predictionExplanations/5..a/",
          "thresholdHigh": 0.835627812,
          "projectId": "578e565a218ef11f13fd2f64",
          "thresholdLow": 0.1,
          "maxExplanations": 3,
          "id": "585be85cdd56a7482bdcb56a",
          "datasetId": "585bbd1dc522f21821107d6e",
          "modelId": "578e59a41ced2e5a9eb18965"
        }
      ],
      "previous": null
    }"""


@pytest.fixture
def pe_json():
    """ PredictionExplanations GET json
    """
    return """{
      "finishTime": 1482424475.33823,
      "numColumns": 50,
      "predictionExplanationsLocation":
        "https://host_name.com/projects/578e...2f64/predictionExplanations/585...9e0f/",
      "thresholdHigh": 2010.0490966796874,
      "projectId": "556cdfbb100d2b0e88585195",
      "thresholdLow": 1706.8041999316963,
      "maxExplanations": 3,
      "id": "585ba071dd56a72ec7109e0f",
      "datasetId": "585b9d71c522f20d38107d6e",
      "modelId": "578e59a41ced2e5a9eb18960"
    }"""


@pytest.fixture
def pe_data(pe_json):
    return from_api(json.loads(pe_json))


@pytest.fixture
def one_pe(pe_data):
    return PredictionExplanations.from_data(pe_data)


@pytest.fixture
def pep_json():
    """ PredictionExplanationsPage GET json
    """
    return """{
      "count": 6,
      "predictionExplanationsRecordLocation":
        "https://host_name.com/projects/5..4/predictionExplanationsRecords/5..7/",
      "next": null,
      "data": [
        {
          "predictionExplanations": [
            {
              "featureValue": 66000,
              "label": "YearMade",
              "strength": 113.18563232421889,
              "feature": "SalePrice",
              "qualitativeStrength": "+++"
            },
            {
              "featureValue": 1139246,
              "label": "YearMade",
              "strength": -84.37227783203139,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            }
          ],
          "prediction": 1843.3448486328125,
          "rowId": 0,
          "predictionValues": [
            {
              "value": 1843.3448486328125,
              "label": "YearMade"
            }
          ]
        },
        {
          "predictionExplanations": [
            {
              "featureValue": 121,
              "label": "YearMade",
              "strength": -140.4251185424805,
              "feature": "datasource",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 57000,
              "label": "YearMade",
              "strength": 107.4373168945308,
              "feature": "SalePrice",
              "qualitativeStrength": "++"
            }
          ],
          "prediction": 1730.043701171875,
          "rowId": 1,
          "predictionValues": [
            {
              "value": 1730.043701171875,
              "label": "YearMade"
            }
          ]
        },
        {
          "predictionExplanations": [
            {
              "featureValue": 1139249,
              "label": "YearMade",
              "strength": -201.79853515624995,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 10000,
              "label": "YearMade",
              "strength": -144.66733398437532,
              "feature": "SalePrice",
              "qualitativeStrength": "---"
            }
          ],
          "prediction": 1751.1256103515625,
          "rowId": 2,
          "predictionValues": [
            {
              "value": 1751.1256103515625,
              "label": "YearMade"
            }
          ]
        },
        {
          "predictionExplanations": [
            {
              "featureValue": 1139251,
              "label": "YearMade",
              "strength": -222.99357910156255,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 1305763200,
              "label": "YearMade",
              "strength": -107.39191894531268,
              "feature": "saledate",
              "qualitativeStrength": "--"
            }
          ],
          "prediction": 1663.2529296875,
          "rowId": 3,
          "predictionValues": [
            {
              "value": 1663.2529296875,
              "label": "YearMade"
            }
          ]
        },
        {
          "predictionExplanations": [
            {
              "featureValue": 1139253,
              "label": "YearMade",
              "strength": -209.14246826171893,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 11000,
              "label": "YearMade",
              "strength": -67.03535156250018,
              "feature": "SalePrice",
              "qualitativeStrength": "--"
            }
          ],
          "prediction": 1724.86328125,
          "rowId": 4,
          "predictionValues": [
            {
              "value": 1724.86328125,
              "label": "YearMade"
            }
          ]
        },
        {
          "predictionExplanations": [
            {
              "featureValue": 1139255,
              "label": "YearMade",
              "strength": -81.31390380859398,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 26500,
              "label": "YearMade",
              "strength": 52.808850097656205,
              "feature": "SalePrice",
              "qualitativeStrength": "++"
            }
          ],
          "prediction": 1881.0067138671875,
          "rowId": 5,
          "predictionValues": [
            {
              "value": 1881.0067138671875,
              "label": "YearMade"
            }
          ]
        }
      ],
      "id": "585bd862dd56a7482bdcb567",
      "previous": null
    }"""


@pytest.fixture
def pep_data(pep_json):
    return from_api(json.loads(pep_json))


@pytest.fixture
def pep_data_with_adjusted_predictions(pep_data):
    pep_data_with_ap = pep_data.copy()
    pep_data_with_ap['adjustment_method'] = 'exposureNormalized'
    for record in pep_data_with_ap['data']:
        # Trivial case, all exposure values set to 1
        record['adjusted_prediction'] = record['prediction']
        record['adjusted_prediction_values'] = record['prediction_values']
    return pep_data_with_ap


@pytest.fixture
def pep_json_classification():
    return """{
      "count": 4,
      "predictionExplanationsRecordLocation":
        "https://host_name.com/projects/5...2/predictionExplanationsRecords/5...c/",
      "next": null,
      "data": [
        {
          "predictionExplanations": [],
          "prediction": 0,
          "rowId": 0,
          "predictionValues": [
            {
              "value": 0.2893982637455831,
              "label": 1
            },
            {
              "value": 0.7106017362544169,
              "label": 0
            }
          ]
        },
        {
          "predictionExplanations": [],
          "prediction": 0,
          "rowId": 1,
          "predictionValues": [
            {
              "value": 0.29246609348399255,
              "label": 1
            },
            {
              "value": 0.7075339065160074,
              "label": 0
            }
          ]
        },
        {
          "predictionExplanations": [],
          "prediction": 0,
          "rowId": 2,
          "predictionValues": [
            {
              "value": 0.3840762929836601,
              "label": 1
            },
            {
              "value": 0.6159237070163399,
              "label": 0
            }
          ]
        },
        {
          "predictionExplanations": [
            {
              "featureValue": 3,
              "strength": -0.3712657301791762,
              "feature": "number_diagnoses",
              "qualitativeStrength": "--",
              "label": 1
            },
            {
              "featureValue": "UN",
              "strength": -0.3112189393477769,
              "feature": "payer_code",
              "qualitativeStrength": "--",
              "label": 1
            },
            {
              "featureValue": "41",
              "strength": -0.2607262089753333,
              "feature": "diag_2",
              "qualitativeStrength": "--",
              "label": 1
            }
          ],
          "prediction": 0,
          "rowId": 3,
          "predictionValues": [
            {
              "value": 0.12064479565777504,
              "label": 1
            },
            {
              "value": 0.8793552043422249,
              "label": 0
            }
          ]
        }
      ],
      "id": "58d1ef9fc80891088925890c",
      "previous": null
    }"""


@pytest.fixture
def pep_data_classification(pep_json_classification):
    return from_api(json.loads(pep_json_classification))


@pytest.fixture
def one_pep(pep_data):
    with mock.patch('datarobot.models.prediction_explanations.t.URL.check',
                    side_effect=(lambda url, context=None: url)):
        return PredictionExplanationsPage.from_data(pep_data)


@pytest.fixture
def rating_table_backend_generated_json():
    return """
    {
        "validationJobId": null,
        "validationError": "",
        "projectId": "556cdfbb100d2b0e88585195",
        "ratingTableName": "custom_table_name1",
        "parentModelId": "5984b4d7100d2b31c1166529",
        "modelJobId": null,
        "id": "5984b4e9100d2b34e5710c9b",
        "originalFilename": "rating_table.csv",
        "modelId": "5984b4d7100d2b31c1166529"
    }"""


@pytest.fixture
def rating_table_backend_generated(rating_table_backend_generated_json):
    return RatingTable.from_server_data(json.loads(rating_table_backend_generated_json))


@pytest.fixture
def rating_table_uploaded_but_not_modeled_json():
    return """
    {
        "validationJobId": "107",
        "validationError": "",
        "projectId": "556cdfbb100d2b0e88585195",
        "ratingTableName": "custom_table_name2",
        "parentModelId": "5983643a100d2b6b89af9622",
        "modelJobId": null,
        "id": "5988b817100d2b62c224691b",
        "originalFilename": "rating_table.csv",
        "modelId": null
    }"""


@pytest.fixture
def rating_table_renamed_json(rating_table_uploaded_but_not_modeled_json):
    rating_table_dict = json.loads(rating_table_uploaded_but_not_modeled_json)
    rating_table_dict['ratingTableName'] = 'renamed'
    return json.dumps(rating_table_dict)


@pytest.fixture
def rating_table_uploaded_but_not_modeled(rating_table_uploaded_but_not_modeled_json):
    return RatingTable.from_server_data(json.loads(
        rating_table_uploaded_but_not_modeled_json))


@pytest.fixture
def renamed_rating_table(rating_table_renamed_json):
    return RatingTable.from_server_data(json.loads(rating_table_renamed_json))


@pytest.fixture
def rating_table_uploaded_and_modeled_json():
    return """
    {
        "validationJobId": "107",
        "validationError": "",
        "projectId": "556cdfbb100d2b0e88585195",
        "ratingTableName": "custom_table_name2",
        "parentModelId": "5983643a100d2b6b89af9622",
        "modelJobId": "114",
        "id": "5988b817100d2b62c224691b",
        "originalFilename": "rating_table.csv",
        "modelId": "5983643a100d2b6b89af9624"
    }"""


@pytest.fixture
def rating_table_uploaded_and_modeled(rating_table_uploaded_and_modeled_json):
    return RatingTable.from_server_data(json.loads(rating_table_uploaded_and_modeled_json))


@pytest.fixture
def blueprint_docs():
    return [{
        'task': u'Approximate kernel support vector classifier.',
        'description': u'Support vector machines are a class of \u201cmaximum margin\u201d ...',
        'links': [
            {'url': u'https://en.wikipedia.org/wiki/Support_vector_machine',
             'name': u'Support vector machine wikipedia'},
            {'url': u'http://scikit-learn.org/stable/modules/svm.html',
             'name': u'Support vector machine scikit-learn'},
            {'url': u'http://scikit-learn.org/stable/modules/kernel_approximation.html',
             'name': u'Kernel Approximation scikit-learn'}],
        'title': u'Nystroem Kernel SVM Classifier Documentation - DataRobot Model Documentation',
        'references': [
            {'url': None,
             'name': u'[R21]Caputo, B., Sim, K., Furesjo, F., & Smola, A. "Appearance-based '
                     u'Object Recognition using SVMs: Which Kernel Should I Use?". In Proc '
                     u'of NIPS workshop on Statistical methods for computational experiments in '
                     u'visual processing and computer vision (2002).'},
            {'url': u'http://bioinformatics.oxfordjournals.org/content/16/10/906.full.pdf',
             'name': u'[R22]Suykens, Johan AK, and Joos Vandewalle. "Least squares '
                     u'support vector machine classifiers." Neural processing '
                     u'letters 9.3 (1999): 293-300. '},
        ],
        'parameters': [
            {'type': u'select (default="nystroem")',
             'name': u'approx (ap)',
             'description': u" The kernel approximation method to use. values: ['nystroem', ..."},
            {'type': u'intgrid (default="100")',
             'name': u'n_components (nc)',
             'description': u' Is the target dimensionality of the feature transform'},
        ]
    }]


@pytest.fixture
def blueprint_chart_data():
    return {
        'nodes': [
            {'id': '1', 'label': 'Task1'},
            {'id': '2', 'label': 'Task2'}
        ],
        'edges': [('1', '2')]
    }


@pytest.fixture
def missing_report_data():
    return {'missing_values_report': [
        {'feature': u'WheelTypeID',
         'type': u'Numeric',
         'missing_count': 136,
         'missing_percentage': 0.0687,
         'tasks': {
             u'2': {'name': 'Missing Values Imputed', 'descriptions': [u'Imputed value: -9999']},
             u'5': {'name': 'Missing Values Imputed', 'descriptions':
                    [u'Missing indicator was treated as feature']}}
         },
        {'feature': u'TopThreeAmericanName',
         'type': u'Categorical',
         'missing_count': 2,
         'missing_percentage': 0.0005,
         'tasks': {u'1': {'name': 'Ordinal encoding of categorical variables',
                          'descriptions': [u'Imputed value: -2']},
                   u'7': {'name': 'One-Hot Encoding',
                          'descriptions': [u'Missing indicator was treated as feature']}}}
    ]}


@pytest.fixture
def invalid_rating_table_json():
    return """
    {
        "validationJobId": "107",
        "validationError": "Error lol",
        "projectId": "556cdfbb100d2b0e88585195",
        "ratingTableName": "custom_table_name2",
        "parentModelId": "5983643a100d2b6b89af9622",
        "modelJobId": "114",
        "id": "5988b817100d2b62c224691b",
        "originalFilename": "rating_table.csv",
        "modelId": "5983643a100d2b6b89af9624"
    }"""


@pytest.fixture
@pytest.mark.usefixtures('known_warning')
def invalid_rating_table(invalid_rating_table_json):
    return RatingTable.from_server_data(json.loads(invalid_rating_table_json))


@pytest.fixture
def multiseries_precomputed_properties(project_url):
    final_payload = {
        'datetimePartitionColumn': 'dates',
        'detectedMultiseriesIdColumns': [
            {
                'multiseriesIdColumns': ['other_eligible'],
                'timeUnit': 'HOUR',
                'timeStep': 5
            }, {
                'multiseriesIdColumns': [
                    'series_id'
                ],
                'timeUnit': 'DAY',
                'timeStep': 1
            }
        ]
    }
    responses.add(responses.GET,
                  '{}features/dates/multiseriesProperties/'.format(project_url),
                  status=200,
                  json=final_payload,
                  content_type='application/json')
    return final_payload['detectedMultiseriesIdColumns'][1]


@pytest.yield_fixture()
def multiseries_postcomputed_properties(project_url):
    initial_payload = {
        'datetimePartitionColumn': 'timestamp',
        'detectedMultiseriesIdColumns': []
    }
    final_payload = {
        'datetimePartitionColumn': 'timestamp',
        'detectedMultiseriesIdColumns': [
            {
                'multiseriesIdColumns': ['other_eligible'],
                'timeUnit': 'HOUR',
                'timeStep': 5
            }, {
                'multiseriesIdColumns': [
                    'series_id'
                ],
                'timeUnit': 'DAY',
                'timeStep': 1
            }
        ]
    }

    with responses.RequestsMock():
        responses.add(responses.GET,
                      '{}features/dates/multiseriesProperties/'.format(project_url),
                      status=200,
                      json=initial_payload,
                      content_type='application/json')
        responses.add(responses.POST,
                      '{}multiseriesProperties/'.format(project_url),
                      body='',
                      status=202,
                      content_type='application/json',
                      adding_headers={'Location': 'https://host_name.com/status/status-id/'})
        responses.add(responses.GET,
                      'https://host_name.com/status/status-id/',
                      status=303,
                      body='',
                      content_type='application/json',
                      adding_headers={
                          'Location': '{}features/dates/multiseriesProperties/'.format(project_url),
                      })
        responses.add(responses.GET,
                      '{}features/dates/multiseriesProperties/'.format(project_url),
                      status=200,
                      json=final_payload,
                      content_type='application/json')
        yield final_payload['detectedMultiseriesIdColumns'][1]


@pytest.yield_fixture
def multiseries_ineligible_properties(project_url):
    final_payload = {
        'datetimePartitionColumn': 'timestamp',
        'detectedMultiseriesIdColumns': []
    }
    with responses.RequestsMock() as req_mock:
        responses.add(responses.POST,
                      '{}multiseriesProperties/'.format(project_url),
                      body='',
                      status=202,
                      content_type='application/json',
                      adding_headers={'Location': 'https://host_name.com/status/status-id/'})
        responses.add(responses.GET,
                      'https://host_name.com/status/status-id/',
                      status=303,
                      body='',
                      content_type='application/json',
                      adding_headers={
                          'Location': '{}features/float/multiseriesProperties/'.format(project_url),
                      })
        responses.add(responses.GET,
                      '{}features/float/multiseriesProperties/'.format(project_url),
                      status=200,
                      json=final_payload,
                      content_type='application/json')
        yield req_mock


@pytest.fixture
def driver_libraries_endpoint(client):
    return '{}/externalDataDriverFile/'.format(client.endpoint)


@pytest.fixture
def drivers_endpoint(client):
    return '{}/externalDataDrivers/'.format(client.endpoint)


@pytest.fixture
def data_stores_endpoint(client):
    return '{}/externalDataStores/'.format(client.endpoint)


@pytest.fixture
def data_sources_endpoint(client):
    return '{}/externalDataSources/'.format(client.endpoint)


@pytest.fixture
def drivers_list_server_resp():
    return {
        'data': [
            {
                'className': 'org.postgresql.Driver',
                'baseNames': ['postgresql-9.4.1208.jre6.jar'],
                'id': '5a99545fe8fc4e0001fa996d',
                'canonicalName': 'PostgreSQL',
                'creator': '5a55dc230936d31663231e3b'
            },
            {
                'className': 'com.mysql.jdbc.Driver',
                'baseNames': ['mysql-connector-java-5.1.44-bin.jar'],
                'id': '5ab37d55438519000121b441',
                'canonicalName': 'mysql',
                'creator': '5ab37d55962d7440474020cc'
            },
            {
                'className': 'com.amazon.redshift.jdbc.Driver',
                'baseNames': ['RedshiftJDBC41-no-awssdk-1.2.10.1009.jar'],
                'id': '5abcac920ca81900010c1535',
                'canonicalName': 'RedShift',
                'creator': '5a55dc230936d31663231e3b'
            }
        ]
    }


@pytest.fixture
def driver_library_upload_server_resp():
    return {'localUrl': 'jdbc_drivers/b6214457-b1f7-403e-8947-60c9aa5710ae.jar'}


@pytest.fixture
def postgresql_driver_server_resp(drivers_list_server_resp):
    return drivers_list_server_resp['data'][0]


@pytest.fixture
def mysql_driver_server_resp(drivers_list_server_resp):
    return drivers_list_server_resp['data'][1]


@pytest.fixture
def redshift_driver_server_resp(drivers_list_server_resp):
    return drivers_list_server_resp['data'][2]


@pytest.fixture
def data_stores_list_server_resp():
    return {
        'data': [
            {
                'updated': '2018-03-25T08:49:05.485000',
                'creator': '5a55dc230936d31663231e3b',
                'canonicalName': 'perftest',
                'params': {
                    'driverId': '5a99545fe8fc4e0001fa996d',
                    'jdbcUrl': 'jdbc:postgresql://bos-dss-borrow1.hq.datarobot.com:5432/perftest'
                },
                'type': 'jdbc',
                'id': '5a995761e8fc4e0001fa996e'
            },
            {
                'updated': '2018-03-25T08:48:55.598000',
                'creator': '5a55dc230936d31663231e3b',
                'canonicalName': 'PostgreSQL Demo',
                'params': {
                    'driverId': '5a99545fe8fc4e0001fa996d',
                    'jdbcUrl': 'jdbc:postgresql://bos-dss-borrow1.hq.datarobot.com:5432/demo'
                },
                'type': 'jdbc',
                'id': '5ab76277bd09db0001f3be16'
            }, {
                'updated': '2018-03-29T09:38:39.107000',
                'creator': '5a55dc230936d31663231e3b',
                'canonicalName': 'RedShift Demo',
                'params': {
                    'driverId': '5abcac920ca81900010c1535',
                    'jdbcUrl': 'jdbc:redshift://10.50.232.182:5439/test'
                },
                'type': 'jdbc',
                'id': '5abcb41f0ca81900010c1536'
            }
        ]
    }


@pytest.fixture
def postgresql_data_store_server_resp(data_stores_list_server_resp):
    return data_stores_list_server_resp['data'][1]


@pytest.fixture
def redshift_data_store_server_resp(data_stores_list_server_resp):
    return data_stores_list_server_resp['data'][2]


@pytest.fixture
def postgresql_schemas_list_server_resp():
    return {
        'catalog': 'perftest',
        'schemas': ['demo', 'information_schema', 'pg_catalog', 'public']
    }


@pytest.fixture
def postgresql_tables_list_server_resp():
    return [
        {
            'type': 'TABLE',
            'name': 'diagnosis',
            'schema': 'demo'
        },
        {
            'type': 'TABLE',
            'name': 'kickcars',
            'schema': 'demo'
        },
        {
            'type': 'TABLE',
            'name': 'patient',
            'schema': 'demo'
        },
        {
            'type': 'TABLE',
            'name': 'transcript',
            'schema': 'demo'
        }
    ]


@pytest.fixture
def data_sources_list_server_resp():
    return {
        'data': [
            {
                'updated': '2018-02-19T12:57:15.365000',
                'creator': '5a530498d5c1f302d6d176c8',
                'canonicalName': 'query_select_all_from_demo_diagnostics',
                'params': {
                    'query': 'SELECT * FROM demo.diagnostics;',
                    'dataStoreId': '5a8ac90b07a57a0001be501e'
                },
                'type': 'jdbc',
                'id': '5a8ac9ab07a57a0001be501f'
            },
            {
                'updated': '2018-04-10T09:29:25.091000',
                'creator': '5a530498d5c1f302d6d176c8',
                'canonicalName': 'Airlines 100mb',
                'params': {
                    'table': 'airlines100mb',
                    'schema': 'public',
                    'dataStoreId': '5acc73e5ec8d670001ba16b2'
                },
                'type': 'jdbc',
                'id': '5acc7420ec8d670001ba16b3'
            }, {
                'updated': '2018-04-10T08:22:40.886000',
                'creator': '5a530498d5c1f302d6d176c8',
                'canonicalName': 'Airlines 10mb',
                'params': {
                    'table': 'airlines10mb',
                    'schema': 'public',
                    'dataStoreId': '5acc73e5ec8d670001ba16b2'
                },
                'type': 'jdbc',
                'id': '5acc7450ec8d670001ba16b4'
            },
            {
                'updated': '2018-04-10T09:30:31.831000',
                'creator': '5a530498d5c1f302d6d176c8',
                'canonicalName': 'Airlines 100mb (1000 records)',
                'params': {
                    'table': 'airlines100mb',
                    'fetchSize': 1000,
                    'dataStoreId': '5acc73e5ec8d670001ba16b2'
                },
                'type': 'jdbc',
                'id': '5acc8437ec8d670001ba16bf'
            }
        ]
    }


@pytest.fixture
def airlines10mb_1000_records_data_source_server_resp(data_sources_list_server_resp):
    return data_sources_list_server_resp['data'][3]


@pytest.fixture
def airlines10mb_data_source_server_resp(data_sources_list_server_resp):
    return data_sources_list_server_resp['data'][2]


@pytest.fixture
def diagnostics_data_source_server_resp(data_sources_list_server_resp):
    return data_sources_list_server_resp['data'][0]



from datetime import datetime

import pytest
import responses
from six.moves.urllib_parse import urlparse, parse_qs

from datarobot import Featurelist, ModelingFeaturelist, Project
from datarobot.utils import from_api, parse_time
from tests.utils import SDKTestcase


class TestFeaturelist(SDKTestcase):

    def test_instantiate_featurelist(self):
        data = {
            'id': '5223deadbeefdeadbeef9999',
            'name': 'Raw Features',
            'features': ['One Fish', 'Two Fish', 'Red Fish', 'Blue Fish'],
            'project_id': '5223deadbeefdeadbeef0101',
            'created': parse_time('2018-07-02T18:49:38.019000Z'),
            'is_user_created': True,
            'num_models': 5
        }

        flist = Featurelist.from_data(data)

        assert flist.id == data['id']
        assert flist.name == data['name']
        assert flist.features == data['features']
        assert repr(flist) == 'Featurelist(Raw Features)'
        assert flist.created == data['created']
        assert flist.is_user_created == data['is_user_created']
        assert flist.num_models == data['num_models']

    @pytest.mark.usefixtures('known_warning')
    def test_instantiate_featurelist_from_dict_deprecated(self):
        data = {
            'id': '5223deadbeefdeadbeef9999',
            'name': 'Raw Features',
            'features': ['One Fish', 'Two Fish', 'Red Fish', 'Blue Fish'],
            'project_id': '5223deadbeefdeadbeef0101',
            'created': parse_time('2018-07-02T18:49:38.019000Z'),
            'is_user_created': True,
            'num_models': 5
        }

        flist = Featurelist(data)

        assert flist.id == data['id']
        assert flist.name == data['name']
        assert flist.features, data['features']
        assert flist.project.id == data['project_id']
        assert repr(flist) == 'Featurelist(Raw Features)'


def test_future_proof(featurelist_server_data):
    Featurelist.from_server_data(dict(featurelist_server_data, future='new'))


@pytest.mark.usefixtures('known_warning')
def test_project_is_known_deprecation(featurelist_server_data):
    fl = Featurelist.from_server_data(dict(featurelist_server_data, future='new'))
    assert fl.project


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_featurelist(featurelist_server_data, project_id):
    url_template = 'https://host_name.com/projects/{}/featurelists/{}/'
    responses.add(responses.GET,
                  url_template.format(project_id, featurelist_server_data['id']),
                  json=featurelist_server_data,
                  status=200,
                  content_type='application/json')
    result = Featurelist.get(project_id, featurelist_server_data['id'])
    assert result.project_id == project_id
    assert result.name == featurelist_server_data['name']
    assert isinstance(result.created, datetime)


@responses.activate
@pytest.mark.usefixtures('known_warning', 'client')
def test_get_featurelist_with_project_instance(featurelist_server_data, project_id):
    responses.add(responses.GET,
                  'https://host_name.com/projects/{}/featurelists/f-id/'.format(project_id),
                  json=featurelist_server_data,
                  status=200,
                  content_type='application/json')

    pdata = {
        'id': project_id,
        'project_name': 'Projects'
    }
    project = Project.from_data(pdata)

    result = Featurelist.get(project, 'f-id')
    assert result.project.id == project_id
    assert result.name == featurelist_server_data['name']


def test_rejects_bad_project_input():
    not_a_project = 5
    with pytest.raises(ValueError):
        Featurelist.get(not_a_project, 'f-id')


def test_print_non_ascii_featurelist(featurelist_server_data):
    hello = u'\u3053\u3093\u306b\u3061\u306f'
    data = dict(featurelist_server_data)
    data['name'] = hello
    featurelist = Featurelist.from_server_data(data)
    print(featurelist)  # actually part of the test - this used to fail (testing __repr__)


@pytest.fixture
def modeling_featurelist_server_data(project_id):
    return {
        'projectId': project_id,
        'id': '5a077ec5a297f9d9229aacaa',
        'name': 'Timeseries Extracted Features',
        'created': '2018-07-02T18:49:38.019000Z',
        'numModels': 5,
        'isUserCreated': False,
        'description': 'A Modeling List',
        'features': [
            'Digital Spend (14 day max)',
            'Digital Spend (14 day mean)',
            'Digital Spend (14 day median)',
            'Digital Spend (14 day min)',
            'Digital Spend (14 day std)',
            'Digital Spend (1st lag)',
            'Digital Spend (21 day max)',
            'Digital Spend (21 day mean)',
            'Digital Spend (21 day median)',
            'Digital Spend (21 day min)',
            'Digital Spend (21 day std)',
            'Digital Spend (28 day max)',
            'Digital Spend (28 day mean)',
            'Digital Spend (28 day median)',
            'Digital Spend (28 day min)',
            'Digital Spend (28 day std)',
            'Digital Spend (2nd lag)',
            'Digital Spend (3rd lag)',
            'Digital Spend (4th lag)',
            'Digital Spend (5th lag)',
            'Digital Spend (7 day max)',
            'Digital Spend (7 day mean)',
            'Digital Spend (7 day median)',
            'Digital Spend (7 day min)',
            'Digital Spend (7 day std)',
            'Forecast Distance',
            'Inventory Rate (14 day max)',
            'Inventory Rate (14 day mean)',
            'Inventory Rate (14 day median)',
            'Inventory Rate (14 day min)',
            'Inventory Rate (14 day std)',
            'Inventory Rate (1st lag)',
            'Inventory Rate (21 day max)',
            'Inventory Rate (21 day mean)',
            'Inventory Rate (21 day median)',
            'Inventory Rate (21 day min)',
            'Inventory Rate (21 day std)',
            'Inventory Rate (28 day max)',
            'Inventory Rate (28 day mean)',
            'Inventory Rate (28 day median)',
            'Inventory Rate (28 day min)',
            'Inventory Rate (28 day std)',
            'Inventory Rate (2nd lag)',
            'Inventory Rate (3rd lag)',
            'Inventory Rate (4th lag)',
            'Inventory Rate (5th lag)',
            'Inventory Rate (7 day max)',
            'Inventory Rate (7 day mean)',
            'Inventory Rate (7 day median)',
            'Inventory Rate (7 day min)',
            'Inventory Rate (7 day std)',
            'Num Employees (14 day max)',
            'Num Employees (14 day mean)',
            'Num Employees (14 day median)',
            'Num Employees (14 day min)',
            'Num Employees (14 day std)',
            'Num Employees (1st lag)',
            'Num Employees (21 day max)',
            'Num Employees (21 day mean)',
            'Num Employees (21 day median)',
            'Num Employees (21 day min)',
            'Num Employees (21 day std)',
            'Num Employees (28 day max)',
            'Num Employees (28 day mean)',
            'Num Employees (28 day median)',
            'Num Employees (28 day min)',
            'Num Employees (28 day std)',
            'Num Employees (2nd lag)',
            'Num Employees (3rd lag)',
            'Num Employees (4th lag)',
            'Num Employees (5th lag)',
            'Num Employees (7 day max)',
            'Num Employees (7 day mean)',
            'Num Employees (7 day median)',
            'Num Employees (7 day min)',
            'Num Employees (7 day std)',
            'Precipitation (14 day max)',
            'Precipitation (14 day mean)',
            'Precipitation (14 day median)',
            'Precipitation (14 day min)',
            'Precipitation (14 day std)',
            'Precipitation (1st lag)',
            'Precipitation (21 day max)',
            'Precipitation (21 day mean)',
            'Precipitation (21 day median)',
            'Precipitation (21 day min)',
            'Precipitation (21 day std)',
            'Precipitation (28 day max)',
            'Precipitation (28 day mean)',
            'Precipitation (28 day median)',
            'Precipitation (28 day min)',
            'Precipitation (28 day std)',
            'Precipitation (2nd lag)',
            'Precipitation (3rd lag)',
            'Precipitation (4th lag)',
            'Precipitation (5th lag)',
            'Precipitation (7 day max)',
            'Precipitation (7 day mean)',
            'Precipitation (7 day median)',
            'Precipitation (7 day min)',
            'Precipitation (7 day std)',
            'Sales (actual)',
            'Sales (log) (7 day diff) (14 day max)',
            'Sales (log) (7 day diff) (14 day mean)',
            'Sales (log) (7 day diff) (14 day median)',
            'Sales (log) (7 day diff) (14 day min)',
            'Sales (log) (7 day diff) (14 day std)',
            'Sales (log) (7 day diff) (1st lag)',
            'Sales (log) (7 day diff) (21 day max)',
            'Sales (log) (7 day diff) (21 day mean)',
            'Sales (log) (7 day diff) (21 day median)',
            'Sales (log) (7 day diff) (21 day std)',
            'Sales (log) (7 day diff) (28 day max)',
            'Sales (log) (7 day diff) (28 day mean)',
            'Sales (log) (7 day diff) (28 day median)',
            'Sales (log) (7 day diff) (28 day min)',
            'Sales (log) (7 day diff) (28 day std)',
            'Sales (log) (7 day diff) (2nd lag)',
            'Sales (log) (7 day diff) (3rd lag)',
            'Sales (log) (7 day diff) (4th lag)',
            'Sales (log) (7 day diff) (5th lag)',
            'Sales (log) (7 day diff) (7 day max)',
            'Sales (log) (7 day diff) (7 day mean)',
            'Sales (log) (7 day diff) (7 day median)',
            'Sales (log) (7 day diff) (7 day min)',
            'Sales (log) (7 day diff) (7 day std)',
            'Sales (log) (naive 7 day seasonal value)',
            'Time (Day of Month) (actual)',
            'Time (Day of Week) (actual)',
            'Time (Month) (actual)',
            'Time (Year) (actual)',
            'Time (actual)',
            'dr_row_type'
        ]
    }


@pytest.fixture
def modeling_featurelists_list_server_data(modeling_featurelist_server_data):
    return {
        'count': 1,
        'next': None,
        'previous': None,
        'data': [modeling_featurelist_server_data]
    }


@pytest.fixture
def modeling_featurelists_with_next_page_server_data(modeling_featurelist_server_data, project_url):
    next_page_url = '{}modelingFeaturelists/?offset=1&limit=1'.format(project_url)
    flist = dict(modeling_featurelist_server_data)
    flist['name'] = 'first_page'
    return {
        'count': 1,
        'next': next_page_url,
        'previous': None,
        'data': [flist]
    }


@pytest.fixture
def modeling_featurelists_with_previous_page_server_data(modeling_featurelist_server_data,
                                                         project_url):
    previous_page_url = '{}modelingFeaturelists/?offset=0&limit=1'.format(project_url)
    flist = dict(modeling_featurelist_server_data)
    flist['name'] = 'second_page'
    return {
        'count': 1,
        'next': None,
        'previous': previous_page_url,
        'data': [flist]
    }


def test_modeling_featurelist_future_proof(modeling_featurelist_server_data):
    future_data = dict(modeling_featurelist_server_data, new='new')
    ModelingFeaturelist.from_server_data(future_data)


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_modeling_featurelist(modeling_featurelist_server_data, project_url, project_id):
    flist_id = modeling_featurelist_server_data['id']
    url = '{}modelingFeaturelists/{}/'.format(project_url, flist_id)
    responses.add(responses.GET, url,
                  json=modeling_featurelist_server_data)

    flist = ModelingFeaturelist.get(project_id, flist_id)
    assert flist.project_id == project_id
    assert flist.name == modeling_featurelist_server_data['name']
    assert flist.features == modeling_featurelist_server_data['features']
    assert flist.id == flist_id
    assert isinstance(flist.created, datetime)
    assert flist.num_models == modeling_featurelist_server_data['numModels']
    assert flist.is_user_created == modeling_featurelist_server_data['isUserCreated']
    assert flist.description == modeling_featurelist_server_data['description']


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_modeling_featurelists(modeling_featurelists_list_server_data, project_url, project):
    responses.add(responses.GET, '{}modelingFeaturelists/'.format(project_url),
                  json=modeling_featurelists_list_server_data)
    flists = project.get_modeling_featurelists()

    assert len(flists) == len(modeling_featurelists_list_server_data['data'])
    assert isinstance(flists[0], ModelingFeaturelist)
    assert flists[0].name == modeling_featurelists_list_server_data['data'][0]['name']


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_modeling_featurelists_paginated(modeling_featurelists_with_next_page_server_data,
                                              modeling_featurelists_with_previous_page_server_data,
                                              project, project_url):
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, '{}modelingFeaturelists/'.format(project_url),
                 json=modeling_featurelists_with_next_page_server_data)
        rsps.add(responses.GET, '{}modelingFeaturelists/'.format(project_url),
                 json=modeling_featurelists_with_previous_page_server_data)
        feats = project.get_modeling_featurelists(batch_size=1)

        first_page_req = rsps.calls[0].request
        second_page_req = rsps.calls[1].request
        assert {'limit': ['1']} == parse_qs(urlparse(first_page_req.url).query)
        assert {'limit': ['1'], 'offset': ['1']} == parse_qs(urlparse(second_page_req.url).query)

    assert len(feats) == 2
    names = {'first_page', 'second_page'}
    assert {feat.name for feat in feats} == names


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_modeling_featurelist(modeling_featurelist_server_data, project_url, project):
    responses.add(responses.POST, '{}modelingFeaturelists/'.format(project_url),
                  json=modeling_featurelist_server_data)

    name = modeling_featurelist_server_data['name']
    features = modeling_featurelist_server_data['features']
    flist = project.create_modeling_featurelist(name, features)

    assert isinstance(flist, ModelingFeaturelist)
    assert flist.id == modeling_featurelist_server_data['id']
    assert flist.name == name
    assert flist.project_id == modeling_featurelist_server_data['projectId']
    assert flist.features == features
    assert flist.is_user_created == modeling_featurelist_server_data['isUserCreated']
    assert flist.num_models == modeling_featurelist_server_data['numModels']
    assert flist.description == modeling_featurelist_server_data['description']
    assert isinstance(flist.created, datetime)


@responses.activate
@pytest.mark.usefixtures('client')
def test_update_featurelist_name_only(featurelist_server_data, project_url):
    featurelist = Featurelist.from_server_data(featurelist_server_data)
    responses.add(responses.PATCH, '{}featurelists/{}/'.format(project_url, featurelist.id),
                  status=204)

    new_name = 'a new name'
    featurelist.update(name=new_name)
    assert featurelist.name == new_name
    assert featurelist.description == featurelist_server_data['description']


@responses.activate
@pytest.mark.usefixtures('client')
def test_update_featurelist_description_only(featurelist_server_data, project_url):
    featurelist = Featurelist.from_server_data(featurelist_server_data)
    responses.add(responses.PATCH, '{}featurelists/{}/'.format(project_url, featurelist.id),
                  status=204)

    new_description = 'a new description'
    featurelist.update(description=new_description)
    assert featurelist.name == featurelist_server_data['name']
    assert featurelist.description == new_description


@responses.activate
@pytest.mark.usefixtures('client')
def test_update_featurelist(featurelist_server_data, project_url):
    featurelist = Featurelist.from_server_data(featurelist_server_data)
    responses.add(responses.PATCH, '{}featurelists/{}/'.format(project_url, featurelist.id),
                  status=204)

    new_name = 'a new name'
    new_description = 'a new description'
    featurelist.update(name=new_name, description=new_description)
    assert featurelist.name == new_name
    assert featurelist.description == new_description


@responses.activate
@pytest.mark.usefixtures('client')
def test_update_modeling_featurelist_name_only(modeling_featurelist_server_data, project_url):
    featurelist = ModelingFeaturelist.from_server_data(modeling_featurelist_server_data)
    old_description = modeling_featurelist_server_data['description']
    responses.add(responses.PATCH, '{}modelingFeaturelists/{}/'.format(project_url, featurelist.id),
                  status=204)

    new_name = 'a new name'
    featurelist.update(name=new_name)
    assert featurelist.name == new_name
    assert featurelist.description == old_description


@responses.activate
@pytest.mark.usefixtures('client')
def test_update_modeling_featurelist_description_only(modeling_featurelist_server_data,
                                                      project_url):
    featurelist = ModelingFeaturelist.from_server_data(modeling_featurelist_server_data)
    responses.add(responses.PATCH, '{}modelingFeaturelists/{}/'.format(project_url, featurelist.id),
                  status=204)

    new_description = 'a new description'
    featurelist.update(description=new_description)
    assert featurelist.name == modeling_featurelist_server_data['name']
    assert featurelist.description == new_description


@responses.activate
@pytest.mark.usefixtures('client')
def test_update_modeling_featurelist(modeling_featurelist_server_data, project_url):
    featurelist = ModelingFeaturelist.from_server_data(modeling_featurelist_server_data)
    responses.add(responses.PATCH, '{}modelingFeaturelists/{}/'.format(project_url, featurelist.id),
                  status=204)

    new_name = 'a new name'
    new_description = 'a new description'
    featurelist.update(name=new_name, description=new_description)
    assert featurelist.name == new_name
    assert featurelist.description == new_description


@responses.activate
@pytest.mark.usefixtures('client')
@pytest.mark.parametrize('dry_run,delete_dependencies',
                         [(False, False), (False, True), (True, False), (True, True)])
def test_delete_featurelist(featurelist_server_data, project_url, dry_run, delete_dependencies):
    featurelist = Featurelist.from_server_data(featurelist_server_data)
    response = {'dryRun': dry_run, 'canDelete': True, 'numAffectedModels': 3, 'numAffectedJobs': 1,
                'deletionBlockedReason': ''}
    responses.add(responses.DELETE,
                  '{}featurelists/{}/'.format(project_url, featurelist.id),
                  status=200, json=response)

    result = featurelist.delete(dry_run=dry_run, delete_dependencies=delete_dependencies)

    delete_request = responses.calls[0].request
    expected_query_params = {'dryRun': [str(dry_run)],
                             'deleteDependencies': [str(delete_dependencies)]}
    assert parse_qs(urlparse(delete_request.url).query) == expected_query_params
    assert result == from_api(response)


@responses.activate
@pytest.mark.usefixtures('client')
def test_delete_featurelist_with_unicode_reason(featurelist_server_data, project_url):
    featurelist = Featurelist.from_server_data(featurelist_server_data)
    response = {'dryRun': False, 'canDelete': True, 'numAffectedModels': 3, 'numAffectedJobs': 1,
                'deletion_blocked_reason': u'\u3053\u3093\u306b\u3061\u306f'}
    responses.add(responses.DELETE,
                  '{}featurelists/{}/'.format(project_url, featurelist.id),
                  status=200, json=response)

    result = featurelist.delete()
    assert result == from_api(response)


@responses.activate
@pytest.mark.usefixtures('client')
@pytest.mark.parametrize('dry_run,delete_dependencies',
                         [(False, False), (False, True), (True, False), (True, True)])
def test_delete_modeling_featurelist(modeling_featurelist_server_data, project_url,
                                     dry_run, delete_dependencies):
    featurelist = ModelingFeaturelist.from_server_data(modeling_featurelist_server_data)
    response = {'dryRun': dry_run, 'canDelete': True, 'numAffectedModels': 3, 'numAffectedJobs': 1,
                'deletionBlockedReason': ''}
    responses.add(responses.DELETE,
                  '{}modelingFeaturelists/{}/'.format(project_url, featurelist.id),
                  status=200, json=response)

    result = featurelist.delete(dry_run=dry_run, delete_dependencies=delete_dependencies)

    delete_request = responses.calls[0].request
    expected_query_params = {'dryRun': [str(dry_run)],
                             'deleteDependencies': [str(delete_dependencies)]}
    assert parse_qs(urlparse(delete_request.url).query) == expected_query_params
    assert result == from_api(response)


@responses.activate
@pytest.mark.usefixtures('client')
def test_delete_modeling_featurelist_with_unicode_reason(modeling_featurelist_server_data,
                                                         project_url):
    featurelist = ModelingFeaturelist.from_server_data(modeling_featurelist_server_data)
    response = {'dryRun': False, 'canDelete': True, 'numAffectedModels': 3, 'numAffectedJobs': 1,
                'deletion_blocked_reason': u'\u3053\u3093\u306b\u3061\u306f'}
    responses.add(responses.DELETE,
                  '{}modelingFeaturelists/{}/'.format(project_url, featurelist.id),
                  status=200, json=response)

    result = featurelist.delete()
    assert result == from_api(response)



from datetime import datetime
import responses

from datarobot import DataSource, DataSourceParameters
from datarobot.utils import parse_time


@responses.activate
def test_list(data_sources_endpoint, data_sources_list_server_resp):
    responses.add(responses.GET, data_sources_endpoint, json=data_sources_list_server_resp)

    data_sources = DataSource.list()

    for data_source, server_payload in zip(data_sources, data_sources_list_server_resp['data']):
        assert data_source.id == server_payload['id']
        assert data_source.type == server_payload['type']
        assert data_source.canonical_name == server_payload['canonicalName']
        assert data_source.creator == server_payload['creator']
        assert data_source.updated == parse_time(server_payload['updated'])
        assert data_source.params.data_store_id == server_payload['params']['dataStoreId']
        # assert params are the same
        params = data_source.params
        srv_params = server_payload['params']
        assert params.table == (srv_params['table'] if 'table' in srv_params else None)
        assert params.schema == (srv_params['schema'] if 'schema' in srv_params else None)
        assert params.partition_column == \
            (srv_params['partitionColumn'] if 'partitionColumn' in srv_params else None)
        assert params.query == (srv_params['query'] if 'query' in srv_params else None)
        assert params.fetch_size == \
            (srv_params['fetchSize'] if 'fetchSize' in srv_params else None)


@responses.activate
def test_get(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source_id = diagnostics_data_source_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp
    )

    data_source = DataSource.get(data_source_id)

    assert data_source.id == diagnostics_data_source_server_resp['id']
    assert data_source.type == diagnostics_data_source_server_resp['type']
    assert data_source.canonical_name == diagnostics_data_source_server_resp['canonicalName']
    assert data_source.creator == diagnostics_data_source_server_resp['creator']
    assert data_source.updated == parse_time(diagnostics_data_source_server_resp['updated'])
    assert data_source.params.data_store_id ==\
        diagnostics_data_source_server_resp['params']['dataStoreId']
    # assert params are the same
    params = data_source.params
    srv_params = diagnostics_data_source_server_resp['params']
    assert params.data_store_id == srv_params['dataStoreId']
    assert params.query == srv_params['query']


@responses.activate
def test_future_proof(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source_id = diagnostics_data_source_server_resp['id']
    aug_resp = dict(diagnostics_data_source_server_resp, newest='newest')
    aug_resp['params']['newKey'] = 'hihihi'
    responses.add(
        responses.GET, '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp
    )

    data_source = DataSource.get(data_source_id)

    assert data_source.id == data_source_id


@responses.activate
def test_create_from_query(data_sources_endpoint, diagnostics_data_source_server_resp):
    responses.add(responses.POST, data_sources_endpoint, json=diagnostics_data_source_server_resp)
    params = DataSourceParameters(
        data_store_id=diagnostics_data_source_server_resp['params']['dataStoreId'],
        query=diagnostics_data_source_server_resp['params']['query']
    )

    data_source = DataSource.create(
        data_source_type=diagnostics_data_source_server_resp['type'],
        canonical_name=diagnostics_data_source_server_resp['canonicalName'],
        params=params
    )

    assert data_source.id == diagnostics_data_source_server_resp['id']
    assert data_source.type == diagnostics_data_source_server_resp['type']
    assert data_source.canonical_name == diagnostics_data_source_server_resp['canonicalName']
    assert data_source.creator == diagnostics_data_source_server_resp['creator']
    assert data_source.updated == parse_time(diagnostics_data_source_server_resp['updated'])
    # assert params are the same
    params = data_source.params
    srv_params = diagnostics_data_source_server_resp['params']
    assert params.data_store_id == srv_params['dataStoreId']
    assert params.query == srv_params['query']
    assert params.partition_column is None
    assert params.table is None
    assert params.schema is None
    assert params.fetch_size is None


@responses.activate
def test_create_from_table(
    data_sources_endpoint, airlines10mb_1000_records_data_source_server_resp
):
    del airlines10mb_1000_records_data_source_server_resp['params']['fetchSize']
    responses.add(responses.POST, data_sources_endpoint,
                  json=airlines10mb_1000_records_data_source_server_resp)
    params = DataSourceParameters(
        data_store_id=airlines10mb_1000_records_data_source_server_resp['params']['dataStoreId'],
        table=airlines10mb_1000_records_data_source_server_resp['params']['table']
    )

    data_source = DataSource.create(
        data_source_type=airlines10mb_1000_records_data_source_server_resp['type'],
        canonical_name=airlines10mb_1000_records_data_source_server_resp['canonicalName'],
        params=params
    )

    assert data_source.id == airlines10mb_1000_records_data_source_server_resp['id']
    assert data_source.type == airlines10mb_1000_records_data_source_server_resp['type']
    assert data_source.canonical_name == \
        airlines10mb_1000_records_data_source_server_resp['canonicalName']
    assert data_source.creator == airlines10mb_1000_records_data_source_server_resp['creator']
    assert data_source.updated == \
        parse_time(airlines10mb_1000_records_data_source_server_resp['updated'])
    # assert params are the same
    params = data_source.params
    srv_params = airlines10mb_1000_records_data_source_server_resp['params']
    assert params.data_store_id == srv_params['dataStoreId']
    assert params.table == srv_params['table']
    assert params.partition_column is None
    assert params.query is None
    assert params.schema is None
    assert params.fetch_size is None


@responses.activate
def test_create_from_table_limited_with_fetch_size(
    data_sources_endpoint, airlines10mb_1000_records_data_source_server_resp
):
    responses.add(responses.POST, data_sources_endpoint,
                  json=airlines10mb_1000_records_data_source_server_resp)
    params = DataSourceParameters(
        data_store_id=airlines10mb_1000_records_data_source_server_resp['params']['dataStoreId'],
        table=airlines10mb_1000_records_data_source_server_resp['params']['table'],
        fetch_size=airlines10mb_1000_records_data_source_server_resp['params']['fetchSize']
    )

    data_source = DataSource.create(
        data_source_type=airlines10mb_1000_records_data_source_server_resp['type'],
        canonical_name=airlines10mb_1000_records_data_source_server_resp['canonicalName'],
        params=params
    )

    assert data_source.id == airlines10mb_1000_records_data_source_server_resp['id']
    assert data_source.type == airlines10mb_1000_records_data_source_server_resp['type']
    assert data_source.canonical_name == \
        airlines10mb_1000_records_data_source_server_resp['canonicalName']
    assert data_source.creator == airlines10mb_1000_records_data_source_server_resp['creator']
    assert data_source.updated == \
        parse_time(airlines10mb_1000_records_data_source_server_resp['updated'])
    # assert params are the same
    params = data_source.params
    srv_params = airlines10mb_1000_records_data_source_server_resp['params']
    assert params.data_store_id == srv_params['dataStoreId']
    assert params.table == srv_params['table']
    assert params.fetch_size == srv_params['fetchSize']
    assert params.partition_column is None
    assert params.query is None
    assert params.schema is None


@responses.activate
def test_create_from_table_specific_schema(
    data_sources_endpoint, airlines10mb_data_source_server_resp
):
    responses.add(responses.POST, data_sources_endpoint,
                  json=airlines10mb_data_source_server_resp)
    params = DataSourceParameters(
        data_store_id=airlines10mb_data_source_server_resp['params']['dataStoreId'],
        table=airlines10mb_data_source_server_resp['params']['table'],
        schema=airlines10mb_data_source_server_resp['params']['schema'],
    )

    data_source = DataSource.create(
        data_source_type=airlines10mb_data_source_server_resp['type'],
        canonical_name=airlines10mb_data_source_server_resp['canonicalName'],
        params=params
    )

    assert data_source.id == airlines10mb_data_source_server_resp['id']
    assert data_source.type == airlines10mb_data_source_server_resp['type']
    assert data_source.canonical_name == \
        airlines10mb_data_source_server_resp['canonicalName']
    assert data_source.creator == airlines10mb_data_source_server_resp['creator']
    assert data_source.updated == \
        parse_time(airlines10mb_data_source_server_resp['updated'])
    # assert params are the same
    params = data_source.params
    srv_params = airlines10mb_data_source_server_resp['params']
    assert params.data_store_id == srv_params['dataStoreId']
    assert params.table == srv_params['table']
    assert params.schema == srv_params['schema']
    assert params.partition_column is None
    assert params.query is None
    assert params.fetch_size is None


@responses.activate
def test_create_from_table_with_partition_column(
    data_sources_endpoint, airlines10mb_data_source_server_resp
):
    del airlines10mb_data_source_server_resp['params']['schema']
    airlines10mb_data_source_server_resp['params']['partitionColumn'] = 'part_column'
    responses.add(responses.POST, data_sources_endpoint,
                  json=airlines10mb_data_source_server_resp)
    params = DataSourceParameters(
        data_store_id=airlines10mb_data_source_server_resp['params']['dataStoreId'],
        table=airlines10mb_data_source_server_resp['params']['table'],
        partition_column=airlines10mb_data_source_server_resp['params']['partitionColumn'],
    )

    data_source = DataSource.create(
        data_source_type=airlines10mb_data_source_server_resp['type'],
        canonical_name=airlines10mb_data_source_server_resp['canonicalName'],
        params=params
    )

    assert data_source.id == airlines10mb_data_source_server_resp['id']
    assert data_source.type == airlines10mb_data_source_server_resp['type']
    assert data_source.canonical_name == \
        airlines10mb_data_source_server_resp['canonicalName']
    assert data_source.creator == airlines10mb_data_source_server_resp['creator']
    assert data_source.updated == \
        parse_time(airlines10mb_data_source_server_resp['updated'])
    # assert params are the same
    params = data_source.params
    srv_params = airlines10mb_data_source_server_resp['params']
    assert params.data_store_id == srv_params['dataStoreId']
    assert params.table == srv_params['table']
    assert params.partition_column == srv_params['partitionColumn']
    assert params.schema is None
    assert params.query is None
    assert params.fetch_size is None


@responses.activate
def test_update_to_table(
    data_sources_endpoint,
    diagnostics_data_source_server_resp,
    airlines10mb_1000_records_data_source_server_resp
):
    data_source_id = diagnostics_data_source_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp)
    updated_params = DataSourceParameters(
        table=airlines10mb_1000_records_data_source_server_resp['params']['table'],
        fetch_size=airlines10mb_1000_records_data_source_server_resp['params']['fetchSize']
    )
    updated_data_source_server_resp = dict(
        diagnostics_data_source_server_resp,
        canonicalName='updated_canonical_name',
        params=updated_params.collect_payload()
    )
    updated_data_source_server_resp['params']['dataStoreId'] = \
        diagnostics_data_source_server_resp['params']['dataStoreId']
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=updated_data_source_server_resp)
    data_source = DataSource.get(data_source_id)

    data_source.update(
        canonical_name=updated_data_source_server_resp['canonicalName'],
        params=updated_params
    )

    assert data_source.id == updated_data_source_server_resp['id']
    assert data_source.type == updated_data_source_server_resp['type']
    assert data_source.canonical_name == updated_data_source_server_resp['canonicalName']
    assert data_source.creator == updated_data_source_server_resp['creator']
    assert data_source.updated == parse_time(updated_data_source_server_resp['updated'])
    assert data_source.params == updated_params


@responses.activate
def test_update_fetch_size(data_sources_endpoint, airlines10mb_data_source_server_resp):
    data_source_id = airlines10mb_data_source_server_resp['id']
    fetch_size = 12345
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp)
    airlines10mb_data_source_server_resp['params']['fetchSize'] = fetch_size
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp)

    data_source = DataSource.get(data_source_id)
    assert data_source.params.fetch_size is None

    new_perams_kwargs = data_source.params.collect_payload()
    new_perams_kwargs['fetch_size'] = fetch_size
    data_source.update(
        params=DataSourceParameters(**new_perams_kwargs)
    )
    new_data_source = DataSource.get(data_source_id)

    assert new_data_source.params.fetch_size == fetch_size


@responses.activate
def test_update_schema(data_sources_endpoint, airlines10mb_data_source_server_resp):
    data_source_id = airlines10mb_data_source_server_resp['id']
    updated_schema = 'other_schema'
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp)
    airlines10mb_data_source_server_resp['params']['schema'] = updated_schema
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp)

    data_source = DataSource.get(data_source_id)
    new_perams_kwargs = data_source.params.collect_payload()
    new_perams_kwargs['schema'] = updated_schema
    data_source.update(
        params=DataSourceParameters(**new_perams_kwargs)
    )
    new_data_source = DataSource.get(data_source_id)
    assert new_data_source.params.schema == updated_schema


@responses.activate
def test_update_partition_column(data_sources_endpoint, airlines10mb_data_source_server_resp):
    data_source_id = airlines10mb_data_source_server_resp['id']
    partition_column = 'my_partition_column'
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp)
    airlines10mb_data_source_server_resp['params']['partition_column'] = partition_column
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_data_source_server_resp)

    data_source = DataSource.get(data_source_id)
    new_perams_kwargs = data_source.params.collect_payload()
    new_perams_kwargs['partition_column'] = partition_column
    data_source.update(
        params=DataSourceParameters(**new_perams_kwargs)
    )
    new_data_source = DataSource.get(data_source_id)
    assert new_data_source.params.partition_column == partition_column


@responses.activate
def test_update_to_query(
    data_sources_endpoint,
    diagnostics_data_source_server_resp,
    airlines10mb_1000_records_data_source_server_resp
):
    data_source_id = airlines10mb_1000_records_data_source_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=airlines10mb_1000_records_data_source_server_resp)
    updated_params = DataSourceParameters(
        query=diagnostics_data_source_server_resp['params']['query'],
    )
    updated_data_source_server_resp = dict(
        airlines10mb_1000_records_data_source_server_resp,
        canonicalName='updated_canonical_name',
        params=updated_params.collect_payload()
    )
    updated_data_source_server_resp['params']['dataStoreId'] = \
        airlines10mb_1000_records_data_source_server_resp['params']['dataStoreId']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=updated_data_source_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=updated_data_source_server_resp)
    data_source = DataSource.get(data_source_id)

    data_source.update(
        canonical_name=updated_data_source_server_resp['canonicalName'],
        params=updated_params
    )

    assert data_source.id == airlines10mb_1000_records_data_source_server_resp['id']
    assert data_source.type == updated_data_source_server_resp['type']
    assert data_source.canonical_name == updated_data_source_server_resp['canonicalName']
    assert data_source.creator == updated_data_source_server_resp['creator']
    assert data_source.updated == parse_time(updated_data_source_server_resp['updated'])
    assert data_source.params == updated_params


@responses.activate
def test_update_unable_to_change_type(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source_id = diagnostics_data_source_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp)

    data_source = DataSource.get(data_source_id)
    data_source._type = 'new_type'
    data_source.update()
    new_data_source = DataSource.get(data_source_id)

    assert new_data_source.type == diagnostics_data_source_server_resp['type']


@responses.activate
def test_update_unable_to_change_creator(
    data_sources_endpoint, diagnostics_data_source_server_resp
):
    data_source_id = diagnostics_data_source_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp)

    data_source = DataSource.get(data_source_id)
    data_source._created = 'other_user'
    data_source.update()
    new_data_source = DataSource.get(data_source_id)

    assert new_data_source.creator == diagnostics_data_source_server_resp['creator']


@responses.activate
def test_update_unable_to_change_updated(
    data_sources_endpoint, diagnostics_data_source_server_resp
):
    data_source_id = diagnostics_data_source_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp)
    responses.add(
        responses.PATCH,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp)

    data_source = DataSource.get(data_source_id)
    data_source._updated = datetime.now()
    data_source.update()
    new_data_source = DataSource.get(data_source_id)

    assert new_data_source.updated == parse_time(diagnostics_data_source_server_resp['updated'])


@responses.activate
def test_delete(data_sources_endpoint, diagnostics_data_source_server_resp):
    data_source_id = diagnostics_data_source_server_resp['id']
    responses.add(
        responses.GET,
        '{}{}/'.format(data_sources_endpoint, data_source_id),
        json=diagnostics_data_source_server_resp
    )
    delete_url = '{}{}/'.format(data_sources_endpoint, data_source_id)
    responses.add(responses.DELETE, delete_url)

    data_source = DataSource.get(data_source_id)
    data_source.delete()

    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == delete_url
